import numpy as np
from scipy.optimize import least_squares
from numpy.fft import rfft, fft, ifft, fftfreq, rfftfreq
import matplotlib.pyplot as plt
from qetpy.utils import shift
from qetpy.core import OFBase

__all__ = ['OF1x1']



class OF1x1:
    """
    Single trace /  single template optimal filter (1x1)
    calculations
    """

    def __init__(self, of_base=None, template_tag='default', 
                 template=None, psd=None, sample_rate=None,
                 pretrigger_msec=None, pretrigger_samples=None,
                 coupling='AC', integralnorm=False,
                 channel_name='unknown',
                 verbose=True):
        
        """
        Initialize OF1x1

        Parameters
        ----------
        
        of_base : OFBase object, optional 
           OF base with pre-calculations
           Default: instantiate base class within OF1x1

        template_tag : str, optional 
           tamplate tag, default='default'
        
        template : ndarray, optional
          template array used for OF calculation, can be 
          None if already in of_base, otherwise required

        psd : ndarray, optional
          psd array used for OF calculation, can be 
          None if already in of_base, otherwise required


        sample_rate : float, optional if of_base is not None
          The sample rate of the data being taken (in Hz)
          Only Required if of_base=None

        pretrigger_samples : int, optional if of_base is not None
            Number of pretrigger samples
            Default: use pretrigger_msec or if also None,
                     use 1/2 trace length

        pretrigger_msec : float, optional if of_base is not None
            Pretrigger length in ms (if  pretrigger_samples is None)
            Default: 1/2 trace length

        coupling : str, optional 
            String that determines if the zero frequency bin of the psd
            should be ignored (i.e. set to infinity) when calculating
            the optimum amplitude. If set to 'AC', then ths zero
            frequency bin is ignored. If set to anything else, then the
            zero frequency bin is kept.
            Default='AC'

        integralnorm : bool, optional
            If set to True, then  template will be normalized 
            to an integral of 1, and any optimum filters will
            instead return the optimum integral in units of Coulombs.
            If set to False, then the usual optimum filter amplitudes
            will be returned (in units of Amps).
            Default=False

        channel_name : str, optional
            channel name
            
        verbose : bool, optional 
            Display information
            Default=True


        Return
        ------
        
        """
    
        # verbose
        self._verbose = verbose

        # tag
        self._template_tag = template_tag

        # Instantiate OF base (if not provided)
        self._of_base = of_base
        if of_base is None:

            # check parameters
            if sample_rate is None:
                raise ValueError('ERROR in OF1x1: sample rate required!')
             
            if (pretrigger_msec is None
                and pretrigger_samples is None):
                raise ValueError('ERROR in OF1x1: '
                                 + 'pretrigger (msec or samples) required!')
                        
            # instantiate
            self._of_base = OFBase(sample_rate, 
                                   pretrigger_msec=pretrigger_msec,
                                   pretrigger_samples=pretrigger_samples,
                                   channel_name=channel_name,
                                   verbose=verbose)
            

        # add template to base object
        if template is not None:

            if self._verbose:
                print('INFO: Adding template with tag "'
                      +  template_tag + '" to OF base object.')
                
            self._of_base.add_template(template,
                                       template_tag=template_tag,
                                       integralnorm=integralnorm)
            
        else:
            
            # check if template exist already
            tags =  self._of_base.template_tags()
        
            if (tags is None
                or template_tag not in tags):

                print('ERROR: No template with tag "'
                      + template_tag + ' found in OF base object.'
                      + ' Modify template tag or add template argument!')
                return
                            
         # add noise to base object
        if psd is not None:

            if self._verbose:
                print('INFO: Adding noise PSD '
                      + 'to OF base object')
            
            self._of_base.set_psd(psd,
                                  coupling=coupling)
            
        else:

            if self._of_base.psd() is None:
                
                print('ERROR: No psd found in OF base object.'
                      + ' Add psd argument!')
                return
        
            
            

        #  template/noise pre-calculation
        if self._of_base.phi(template_tag) is None:
            self._of_base.calc_phi(template_tags=template_tag)



        # initialize fit results

        # single 1x1
        self._of_amp_nodelay = None
        self._of_chisq_nodelay = None
        self._of_t0_nodelay = None
        self._of_chi2low_nodelay = None

        self._of_amp_withdelay = None
        self._of_chisq_withdelay = None
        self._of_t0_withdelay = None
        self._of_chi2low_withdelay = None
        

        # iterative
        self._of_amps_iterative = None
        self._of_chisq_iterative = None
        self._of_t0_iterative = None
        self._of_chi2low_iterative = None


            
    def calc(self, signal=None, 
             window_min_from_trig_usec=None,
             window_max_from_trig_usec=None,
             window_min_index=None,
             window_max_index=None,
             lowchi2_fcutoff=10000,
             lgc_outside_window=False,
             pulse_direction_constraint=0,
             interpolate_t0=False,
             lgc_fit_nodelay=True,
             lgc_plot=False):
        """
        Calculate OF with delay (and no delay if 
        lgc_fit_nodelay=True)

        Parameters
        ----------
        
        signal : ndarray, optional
          signal trace, can be None if already
          set in 'of_base
          
        window_min_from_trig_usec : float, optional
           OF filter window start in micro seconds from
           pre-trigger (can be negative if prior pre-trigger)

        window_max_from_trig_usec : float, optional
           OF filter window end in micro seconds from
           pre-trigger (can be negative if prior pre-trigger)


        window_min_index: int, optional
            OF filter window start in ADC samples 
        
        window_max_index: int, optional
            OF filter window end in ADC samples
            
        lowchi2_fcutoff : float, optional
            The frequency (in Hz) that we should cut off the chi^2 when
            calculating the low frequency chi^2. Default is 10 kHz.
            
        lgcoutsidewindow : bool, optional
            Boolean flag that is used to specify whether the Optimum
            Filter should look inside `nconstrain` or outside it. If
            False, the filter will minimize the chi^2 in the bins
            specified by `nconstrain`, which is the default behavior.
            If True, then it will minimize the chi^2 in the bins that
            do not contain the constrained window.


        pulse_direction_constraint : int, optional
            Sets a constraint on the direction of the fitted pulse.
            If 0, then no constraint on the pulse direction is set.
            If 1, then a positive pulse constraint is set for all fits.
            If -1, then a negative pulse constraint is set for all
            fits. If any other value, then a ValueError will be raised.
      
        interpolate_t0 : bool, optional
            If True, then a precomputed solution to the parabolic
            equation is used to find the interpolated time-of-best-fit.
            Default is False.

        lgc_fit_nodelay : bool, option
            calculation no-delay OF, default=True

        lgc_plot : bool, optional
            display diagnostic plot

        Return
        ------
        None

        """

        
        # update signal and do preliminary
        # calculations
        if signal is not None:   
            self._of_base.update_signal(
                signal,
                calc_signal_filt=True,
                calc_signal_filt_td=True,
                calc_chisq_amp=True,
                template_tags=self._template_tag
            )
            
        # get fit results
        amp,t0,chisq = self._of_base.get_fit_withdelay(
            self._template_tag,
            window_min_from_trig_usec=window_min_from_trig_usec,
            window_max_from_trig_usec=window_max_from_trig_usec,
            window_min_index=window_min_index,
            window_max_index=window_max_index,
            lgc_outside_window=lgc_outside_window,
            pulse_direction_constraint=pulse_direction_constraint,
            interpolate_t0=interpolate_t0
        )
        
        lowchi2 = self._of_base.get_chisq_lowfreq(
            template_tag=self._template_tag,
            amp=amp,
            t0=t0,
            lowchi2_fcutoff=lowchi2_fcutoff
        )

        self._of_amp_withdelay = amp
        self._of_chisq_withdelay = chisq
        self._of_t0_withdelay = t0
        self._of_chi2low_withdelay = lowchi2
        
        # add nodelay fit
        if lgc_fit_nodelay:
            amp_0,t0_0,chisq_0 = self._of_base.get_fit_nodelay(
                template_tag=self._template_tag,
                shift_usec=None,
                use_chisq_alltimes=True
            )
            lowchi2_0 = self._of_base.get_chisq_lowfreq(
                template_tag=self._template_tag,
                amp=amp_0,
                t0=t0_0,
                lowchi2_fcutoff=lowchi2_fcutoff
            )
        
            self._of_amp_nodelay = amp_0
            self._of_chisq_nodelay = chisq_0
            self._of_t0_nodelay = t0_0
            self._of_chi2low_nodelay = lowchi2_0

        if lgc_plot:
            self.plot(lgc_plot_withdelay=True,
                      lgc_plot_nodelay=lgc_fit_nodelay)
                
                
                

    def calc_nodelay(self, signal=None,
                     shift_usec=None,
                     lowchi2_fcutoff=10000,
                     use_chisq_alltimes=True,
                     lgc_plot=False):
        """
        Calculate no-delay OF  (a shift
        from pretrigger can be added)

        Parameters
        ----------

        signal : ndarray, optional
          signal trace, can be None if already
          set in 'of_base

        shift_usec : float, optional
          shift in micro seconds from pretrigger time
          default: no shift
                             
        lowchi2_fcutoff : float, optional
            The frequency (in Hz) that we should cut off the chi^2 when
            calculating the low frequency chi^2. Default is 10 kHz.
        
        use_chisq_alltimes : bool, optional
          use the chisq all times 
        
        lgc_plot : bool, optional
            display diagnostic plot


        Return
        ------
        None

        """


        # update signal and do preliminary
        # calculations
        if signal is not None:   
            self._of_base.update_signal(
                signal,
                calc_signal_filt=True,
                calc_signal_filt_td=True,
                calc_chisq_amp=True,
                template_tags=self._template_tag
            )
        
        # nodelay fit
        amp_0,t0_0,chisq_0 = self._of_base.get_fit_nodelay(
            template_tag=self._template_tag,
            shift_usec=shift_usec,
            use_chisq_alltimes=use_chisq_alltimes
        )
        
        lowchi2_0 = self._of_base.get_chisq_lowfreq(
            template_tag=self._template_tag,
            amp=amp_0,
            t0=t0_0,
            lowchi2_fcutoff=lowchi2_fcutoff
        )
        
        self._of_amp_nodelay = amp_0
        self._of_chisq_nodelay = chisq_0
        self._of_t0_nodelay = t0_0
        self._of_chi2low_nodelay = lowchi2_0

        if lgc_plot:
            self.plot(lgc_fit_nodelay=True)
            

        

    def get_result_nodelay(self):
        """
        Get OF no-delay results

        Parameters
        ----------
        none

        Return
        ------
  
        amp : float
            The optimum amplitude calculated for the trace (in Amps)
            with no time shifting allowed (or at the time specified by
            'shift_usec').

         t0 : float
            The time shift (=0 or shift_usec)

        chisq : float
            The chi^2 value calculated from the optimum filter with no
            time shifting (or at the time shift specified by shift_usec)
            
        lowchi2_0 : float
            The low frequency chi^2 value (cut off at lowchi2_fcutoff) for the
            inputted values with no time shifting (or at the time shift 
            specified by shift_usec)
        
        """
        
        return (self._of_amp_nodelay,
                self._of_t0_nodelay,
                self._of_chisq_nodelay, 
                self._of_chi2low_nodelay)


    def get_result_withdelay(self):
        """
        Get OF with delay results
 
        Parameters
        ----------
        none

        Return
        ------

        amp : float
            The optimum amplitude calculated for the trace (in Amps).
        t0 : float
            The time shift calculated for the pulse (in s).
        chi2 : float
            The chi^2 value calculated from the optimum filter.
        lowchi2 : float
            The low frequency chi^2 value (cut off at lowchi2_fcutoff) for the
            inputted values
        """

        
        return (self._of_amp_withdelay,
                self._of_t0_withdelay,
                self._of_chisq_withdelay, 
                self._of_chi2low_withdelay)


    
    def get_energy_resolution(self):
        """
        Method to return the energy resolution for the optimum filter.
        (resolution depends only on template and noise!) 
        
        Parameters
        ----------
        None


        Returns
        -------
        sigma : float
            The energy resolution of the optimum filter.

        """
        sigma = self._of_base.get_energy_resolution(
            self._template_tag
        )

        return sigma

    
    def get_time_resolution(self, amp):
        """
        Method to return the time resolution for the optimum filter.
        Resolution depends also on fitted amplitude (-> reset every events)
 
        Parameters
        ----------

        amp : float
          OF fitted amplitude
       

        Returns
        -------
        sigma : float
            The time resolution of the optimum filter.

        """

        sigma =  self._of_base.get_time_resolution(
            amp,
            self._template_tag
        )

        return sigma



    
    def plot(self, lgc_plot_withdelay=True,
             lgc_plot_nodelay=True,
             figsize=(8, 5),
             xlim_msec=None):
        """
        Diagnostic plot

        Parameters
        ----------
        lgc_plot_withdelay : bool, optional
           If true, include OF with delay
           Default=True

        lgc_plot_nodelay : bool, optional
           If true, include OF no-delay
           Default=True

        figsize : tuple, optional
           figure size
           Default=(8, 5)

        xlim_msec : array like, optional
          min/max x-axis
        
        Return
        ------
        none
        
        """


        # check
        if lgc_plot_withdelay and self._of_amp_withdelay is None:
            print('ERROR: No fit (with delay) done. Unable to plot result!')
            return

        if lgc_plot_nodelay and self._of_amp_nodelay is None:
            print('ERROR: No fit (no delay) done. Unable to plot result!')
            return
        
        # signal
        signal = self._of_base.signal()
        template = self._of_base.template()
        fs = self._of_base.sample_rate
        nbins = len(signal)
        chi2 = self._of_chisq_withdelay/len(signal)
        
        # time axis
        xtime_ms = 1e3*np.arange(nbins)/fs


        
        # define figure abd plot
        fig, ax = plt.subplots(figsize=figsize)   
        ax.plot(xtime_ms, signal*1e6, label='Signal', color='blue', alpha=0.5)

        if lgc_plot_withdelay:
            chi2 = self._of_chisq_withdelay/len(signal)
            
            ax.plot(xtime_ms,
                    self._of_amp_withdelay*np.roll(
                        template,
                        int(self._of_t0_withdelay*fs)
                    )*1e6, 
                    label=(r'OF with delay, $\chi^2$'
                           + f'/Ndof={chi2:.2f}'),
                    color='red',
                    linestyle='dotted')

        if lgc_plot_nodelay:
            chi2 = self._of_chisq_nodelay/len(signal)
            ax.plot(xtime_ms,
                    self._of_amp_nodelay*np.roll(
                        template,
                        int(self._of_t0_nodelay*fs)
                    )*1e6, 
                    label=(r'OF no delay, $\chi^2$'
                           + f'/Ndof={chi2:.2f}'),
                    color='green',
                    linestyle='dotted')
            


        if xlim_msec is not None:
            ax.set_xlim(xlim_msec)
        ax.set_ylabel(r'Current [$\mu A$]')
        ax.set_xlabel('Time [ms]')
        ax.set_title(f'{self._of_base.channel_name} OF Results')
        lgd = ax.legend(loc='best')
        ax.tick_params(which='both', direction='in', right=True, top=True)
        ax.grid(linestyle='dotted')
        fig.tight_layout()



import numpy as np
from math import ceil, floor
from numpy.fft import rfft, fft, ifft, fftfreq, rfftfreq
from qetpy.utils import shift, interpolate_of, argmin_chisq


__all__ = ['OFBase']


class OFBase:
    """
    Single channel (trace) - multiple templates,
    optimal filter base class

    The calculation use single noise psd, single pre-trigger
    samples
    
    """

    def __init__(self, sample_rate,
                 pretrigger_samples=None,
                 pretrigger_msec=None,
                 channel_name='unknown',
                 verbose=True):
        """
        Initialization of the optimum filter base class
        
        Parameters
        ----------

        sample_rate : float
            The sample rate of the data being taken (in Hz).

        
        pretrigger_samples : int, optional
            Number of pretrigger samples
            Default: use pretrigger_msec or if also None,
                     use 1/2 trace length

        pretrigger_msec : float, optional
            Pretrigger length in ms (if  pretrigger_samples is None)
            Default: 1/2 trace length


        channel_name : str, optional
            Name of the channel 
            Default='unknown' 

        verbose : bool, optional 
            Display information
            Default=True


        Return
        ------
        None

        """

        
        self._debug = False
        self._verbose = verbose
        self._channel_name = channel_name
        self._fs = sample_rate
        self._pretrigger_samples = pretrigger_samples
        if pretrigger_msec is not None:
            self._pretrigger_samples = (
                int(round(pretrigger_msec*1e-3*self._fs))
            )
            

        # Initialize some parameters
            
        # initialize nb samples 
        self._nbins = None

        # initialize frequency spacing of FFT and frequencies
        self._df = None
        self._fft_freqs = None

        
        # initialize templates (time domain and FFT)
        # dict key = template tag
        self._templates = dict()
        self._templates_fft = dict()

        # initialize two-sided noise psd (in Amps^2/Hz)
        self._psd = None
        self._psd_tag = None

        # initialize calculated optimal filter abd norm
        # (not dependent of signal)
        # dict key = template tag
        self._phis = dict()
        self._norms = dict()
            
        # initialize signal
        self._signal = None
        self._signal_fft = None
    
        # initialize (optimal) filtered  signals and templates
        # (frequency domain and  converted back to time domain)
        # dict key = template tag
        self._signal_filts = dict()
        self._signal_filts_td = dict()
        self._template_filts = dict()
        self._template_filts_td = dict()


        # initialize amplitudes and chi2 (for all times)
        # dict key = template tag
        self._chisq0 = None # "no pulse" chisq (independent of template)
        self._chisqs_alltimes_rolled = dict() # chisq all times
        self._amps_alltimes_rolled = dict() # amps all times
     
               
        if self._debug:
            print('DEBUG: Instantiate OF base for channel '
                  + channel_name)


        
    @property
    def verbose(self):
        return self._verbose

    @property
    def channel_name(self):
        return self._channel_name


    @property
    def sample_rate(self):
        return self._fs

    @property
    def pretrigger_samples(self):
        return self._pretrigger_samples


    


    def template_tags(self):
        """
        get template tags

        Parameters
        ----------
        None

        Return
        ------
        
        tag : list
         list with template tags
        
        """
        
        if self._templates:
            return list(self._templates.keys())
        else:
            return []


    
    def template(self, template_tag='default'):
        """
        Get template in time domain for the specified tag

        Parameters
        ----------
        
        template_tag : str, optional
          template tag/id
          default: 'default'
 

        Return
        ------
        
        template : ndarray
         template trace in time domain

        """
        
        if  template_tag in self._templates.keys():
            return self._templates[template_tag]
        else:
            return None


        

    def template_fft(self, template_tag='default'):
        """
        Get template FFT for the specified tag

        Parameters
        ----------
        
        template_tag : str, optional
          template tag/id
          default: 'default'
 

        Return
        ------
        
        template : ndarray
         template trace FFT

        """
        
        if  template_tag in self._templates_fft.keys():
            return self._templates_fft[template_tag]
        else:
            return None    

        
    def psd(self):
        """
        Get psd

        Parameters
        ----------        
        None


        Return
        ------
        
        psd : ndarray
         noise PSD

        """

        return self._psd


    def psd_tag(self):
        """
        Get psd tag
        
        Parameters
        ----------        
        None


        Return
        ------
        psd_tag: str
         PSD tag 
        """

        return self._psd_tag


    def signal(self):
        """
        Get current signal trace in time domain
        
        
        Parameters
        ----------        
        None


        Return
        ------
        signal : ndarray
          time domain signal trace    

        """

        return self._signal


    
    def signal_fft(self):
        """
        Get current signal trace FFT
          
        Parameters
        ----------        
        None


        Return
        ------
        signal_fft : ndarray
          signal trace FFT    

        """

        return self._signal_fft

    

    def phi(self, template_tag='default'):
        """
        Get "optimal filter" (phi) for a specified
        template tag, depends only
        on template and PSD

        Parameters
        ----------
        
        template_tag : str, optional
          template tag/id
          default: 'default;='

        Return
        ------
        
        phi : ndarray
          "optimal filter" values

        """
        
        if template_tag in self._phis.keys():
            return self._phis[template_tag]
        else:
            return None


        
    def norm(self, template_tag='default'):
                 
        """
        Method to return norm for the optimum filter
        (this is the denominator of amplitude estimate)
        for the specified template tag

        Parameters
        ----------
        
        template_tag : str, optional
          template tag/id
          default: 'default'


        Returns
        -------
        norm : float
            normalization for the optimum filter
        """
           
        if template_tag in self._norms.keys():
            return self._norms[template_tag]
        else:
            return None 

        
    
    def signal_filt(self, template_tag='default'):
        """
        Get (optimal) filtered signal in frequency domain
        for the specified template tag

        signal_filt = phi*signal_fft/norm
 
        Parameters
        ----------
        
        template_tag : str, optional
          template tag/id
          default: 'default'


        Returns
        -------
        signal_filt  :ndarray
           optimal filtered signal 

        """

        if template_tag in self._signal_filts.keys():
            return self._signal_filts[template_tag]
        else:
            return None

        
    def signal_filt_td(self, template_tag='default'):
        """
        Get (optimal) filtered signal converted back to time domain
        for the specified template tag

        signal_filt_td = ifft(signal_filt)
                       = ifft(phi*signal_fft/norm)
 
        Parameters
        ----------
        
        template_tag : str, optional
          template tag/id
          default: 'default'


        Returns
        -------
        signal_filt_td  :ndarray
           optimal filtered signal in time domain

        """

        if template_tag in self._signal_filts_td.keys():
            return self._signal_filts_td[template_tag]
        else:
            return None

  


        
    def template_filt(self, template_tag='default'):
        """
        FIXME: no implemented yet

        Get (optimal) filtered template in frequency domain
        for a specified template tag
        
        Parameters
        ----------
        
        template_tag : str, optional
          template tag/id
          default: 'default'


        Returns
        -------
        template_filt  :ndarray
           optimal filtered template in fourier domain
        

        """

        if template_tag in self._template_filts.keys():
            return self._template_filts[template_tag]
        else:
            return None
        


        
    def template_filt_td(self, template_tag='default'):
        """
        FIXME: no implemented yet

        Get (optimal) filtered template converted back to time domain
        for a specified template tag
        
        Parameters
        ----------
        
        template_tag : str, optional
          template tag/id
          default: 'default'


        Returns
        -------
        template_filt_td  :ndarray
           optimal filtered template in time domain
        


        """

        if template_tag in self._template_filts_td.keys():
            return self._template_filts_td[template_tag]
        else:
            return None
       

        

    def add_template(self, template, template_tag='default',
                     integralnorm=False):
        """
        Add template with a user specified tag. If template
        with same tag already exist, it is overwritten!

        immediately calculate template FFT automatically 
        
        Parameters
        ----------
        
        template : ndarray 
           template numpy 1d array
        
        template_tag : string, optional [default='default']
           name associated to the template
        
        integralnorm : bool, optional [default = False]
            If set to True, then  template will be normalized 
            to an integral of 1, and any optimum filters will
            instead return the optimum integral in units of Coulombs.
            If set to False, then the usual optimum filter amplitudes
            will be returned (in units of Amps).


        Returns
        -------
        None

        """

        # add to dictionary
        self._templates[template_tag] = template

        # FFT
        if  self._nbins is None:
            self._nbins = template.shape[0]
            self._df = self._fs/self._nbins
            self._fft_freqs = fftfreq(self._nbins, d=1.0/self._fs)
            
        elif template.shape[0]!=self._nbins:
            raise ValueError('Inconsistent number of samples')

        self._templates_fft[template_tag] = fft(template, axis=-1)/self._nbins/self._df
      
        
        if integralnorm:
            self._templates_fft[template_tag]  /= self._templates_fft[template_tag][0]

        # pre-trigger
        if self._pretrigger_samples  is None:
            self._pretrigger_samples = self._nbins//2

        # debug
        if self._debug:
            print('DEBUG: Add template "'
                  + template_tag + '"')

            
            
    def set_psd(self, psd, coupling="AC", psd_tag='default'):
        """
        Add psd, a tag can be specified 
        If psd already exist, it is overwritten and tag replaced.
        
        
        Parameters
        ----------
        
        psd : ndarray 
           psd 1d array
                
        coupling : str, optional [default='AC']
            String that determines if the zero frequency bin of the psd
            should be ignored (i.e. set to infinity) when calculating
            the optimum amplitude. If set to 'AC', then ths zero
            frequency bin is ignored. If set to anything else, then the
            zero frequency bin is kept. 

        Returns
        -------
        None

        """
            
        
        # add to dictionary
        self._psd  = np.zeros(len(psd))
        self._psd[:] = psd

        # check coupling
        if coupling=="AC":
            self._psd[0] = np.inf


        # tag
        self._psd_tag = psd_tag
      


        
    def clear_signal(self):
        """
        Method to intialize calculated signal
        parameters
 
        Parameters
        ----------
        None

        Return
        ---------
        None

        """

        # signal
        self._signal = None
        self._signal_fft = None
    
        # (optimal) filtered  signals and templates
        # (frequency domain and  converted back to time domain)
        self._signal_filts = dict()
        self._signal_filts_td = dict()
        self._template_filts = dict()
        self._template_filts_td = dict()


        # chisq and amp arrays
        self._chisq0 = None
        self._chisqs_alltimes_rolled = dict()
        self._amps_alltimes_rolled = dict()
        
    
            
    def update_signal(self, signal,
                      calc_signal_filt=True,
                      calc_signal_filt_td=True,
                      calc_chisq_amp=True,
                      template_tags=None):
        """
        Method to update new signal trace, needs to be called each event

        Parameters
        ----------
        
        signal : ndarray 
           the signal that we want to apply the optimum filter to
           (units should be Amps).

        calc_signal_filt : bool, optional
           If true calculate signal filt
           (for tags specified with "template_tags" or all tags if None)
           Default: True

        
        calc_signal_filt_td : bool, optional
           If true calculate signal filt and convert back to time domain
           (for tags specified with "template_tags" or all tags if None)
           Default: True

        calc_chisq_amp : bool, optional
           If true calculate (rolled) chisq/amps for all times
           (for tags specified with "template_tags" or all tags if None)
           Default: True

        template_tags : list 
         list of template tags for the above calculations
         Dafault: all tags 
               
         
        Return
        ------
        None


        """

        # check nb samples
        if signal.shape[0]!=self._nbins:
            raise ValueError('Inconsistent number of samples '
                             + 'between signal and template')


        # reset all signal dependent quantities
        self.clear_signal()


        # debug
        if self._debug:
            print('DEBUG: Update signal for channel "'
                  + self._channel_name + '"!')
            
            
        
        # update signal
        self._signal = signal

        # FFT
        self._signal_fft = fft(signal, axis=-1)/self._nbins/self._df
       
        
        if calc_signal_filt or calc_signal_filt_td:
            
            # calculate filtered signal
            self.calc_signal_filt(template_tags=template_tags)
            
            # calc filtered signal time domain
            if calc_signal_filt_td:
                self.calc_signal_filt_td(template_tags=template_tags)

                
        # calc chisq no pulse 
        if calc_chisq_amp:
            self.calc_chisq_amp(template_tags=template_tags)

     

    def calc_phi(self, template_tags=None):
        """
        calculate optimal filters (phi)

        phi = template_fft* / psd
        
        Parameters
        ----------
        template_tags : NoneType or str or list of string 
                        [default=None]    
           template tags to calculate optimal filters, if None, 
           calculate optimal filter for all templates

        Return
        ------
        None
        """

        if template_tags is None:
            template_tags = self._templates_fft.keys()
        elif isinstance(template_tags, str):
            template_tags = [template_tags]
        elif not isinstance(template_tags, list):
            raise ValueError('"template_tags argument should be a '
                             + ' a string or list of strings')


        # loop and calculate optimal filters
        for tag in template_tags:

            if (tag not in self._templates_fft.keys()
                or self._psd is None):
                raise ValueError('Template or psd tag not found!')


            if self._debug:
                print('DEBUG: Calculating phi with template "'+
                      tag + '"')
            
            # calculate OF
            self._phis[tag] = (
                self._templates_fft[tag].conjugate() / self._psd
            )
            
            
            # calculate norm
            self._norms[tag] = (
                np.real(np.dot(self._phis[tag],
                               self._templates_fft[tag]))*self._df
            )

        

    def calc_signal_filt(self, template_tags=None):
        """
        Calculate filtered signal (for the specified
        or all template tags)
 
        signal_filt = phi*signal_fft/norm
         
        Parameters
        ----------
        template_tags : NoneType or str or list of string 
                        [default=None]    
           template tags to calculate optimal filters, if None, 
           calculate optimal filter for all templates

        Return
        ------
        None

        """


        # check if phis have been calculcae
        if not self._phis:
            self.calc_phi(template_tags=template_tags)

        
        if template_tags is None:
            template_tags = self._phis.keys()
        elif isinstance(template_tags, str):
            template_tags = [template_tags]
        elif not isinstance(template_tags, list):
            raise ValueError('"template_tags argument should be a '
                             + ' a string or list of strings')

        for tag in template_tags:
            
            if tag not in self._phis.keys():
                self.calc_phi(template_tags=tag)
                
            # filtered signal
            self._signal_filts[tag] = (
                self._phis[tag]*self._signal_fft/self._norms[tag]
            )

            # debug
            if self._debug:
                print('DEBUG: Calculating signal_filt with template "'+
                      tag + '"')

    def calc_signal_filt_td(self, template_tags=None):
        """
        Convert signal filt to time domain (for the specified
        or all template tags)

        signal_filt_td = ifft(signal_filt)
                       = ifft(phi*signal_fft/norm)


        Parameters
        ----------
        template_tags : NoneType or str or list of string 
                        [default=None]    
           template tags to calculate optimal filters, if None, 
           calculate optimal filter for all templates

        Return
        ------
        None

        """

        # check if filtered signal available
        if not self._signal_filts:
            self.calc_signal_filt(
                template_tags=template_tags
            )

        # get tags
        if template_tags is None:
            template_tags = self._signal_filts.keys()
        elif isinstance(template_tags, str):
            template_tags = [template_tags]
        elif not isinstance(template_tags, list):
            raise ValueError('"template_tags argument should be a '
                             + ' a string or list of strings')

        
        for tag in template_tags:

            # calc signal filt ifft
            self._signal_filts_td[tag] = np.real(
                ifft(self._signal_filts[tag]*self._nbins, axis=-1)
            )*self._df
            
            # debug
            if self._debug:
                print('DEBUG: Calculating signal_filt_td with template "'+
                      tag + '"')
            
    def calc_chisq0(self):
        """
        Calculate part of chi2 that doesn't depend 
        on template (aka "no pulse chisq)

        Parameters
        ----------
        None


        Return
        ------
        None
        
        """

        # "no pulse chisq" (doesn't depend on template)
        self._chisq0 = np.real(
            np.dot(self._signal_fft.conjugate()/self._psd,
                   self._signal_fft)*self._df
        )


        
    def calc_chisq_amp(self, template_tags=None):
        """
        Calculate chi2/amp for all times (rolled
        so that 0-delay is the pretrigger bin)

        Parameters
        ----------
        template_tags : NoneType or str or list of string 
                        [default=None]    
           template tags to calculate optimal filters, if None, 
           calculate optimal filter for all templates

        Return
        ------
        None
        

        """
        
        # "no pulse chisq" (doesn't depend on template)
        self.calc_chisq0()
        
        
        # time dependent chisq + sum of the two

        # check if filtered signal (ifft) available
        # if not calculate
        if not self._signal_filts_td:
            self.calc_signal_filt_td(
                template_tags=template_tags
            )
            

        # find tags
        if template_tags is None:
            template_tags = self._signal_filts_td.keys()
        elif isinstance(template_tags, str):
            template_tags = [template_tags]
        elif not isinstance(template_tags, list):
            raise ValueError('"template_tags argument should be a '
                             + ' a string or list of strings')


                  
        # loop tags
        for tag in template_tags:

            # build chi2
            chisq_t0 = (self._signal_filts_td[tag]**2)*self._norms[tag]
            
            # total chisq
            chisq = self._chisq0 - chisq_t0

        
            # shift so that 0 delay is at pre-trigger bin
            chisq_rolled = np.roll(chisq,
                                   self._pretrigger_samples,
                                   axis=-1)

            self._chisqs_alltimes_rolled[tag] = chisq_rolled

            
            # amplitude
            self._amps_alltimes_rolled[tag] = np.roll(self._signal_filts_td[tag],
                                                      self._pretrigger_samples,
                                                      axis=-1)
            
            # debug
            if self._debug:
                print('DEBUG: Calculating chisq/amp all times with template "'+
                      tag + '"')
          
        
    def get_fit_nodelay(self, template_tag='default',
                        shift_usec=None,
                        use_chisq_alltimes=True):
        """
        Function to calculat and return the optimum amplitude/chisq of a pulse in
        data with no time shifting, or at a specific time.
        
        Parameters
        ----------

        template_tag : str, optional
          template tag/id
          default: 'default'

        shift_usec : float, optional
          shift in micro seconds from pretrigger time
          default: no shift
        
        use_chisq_alltimes : bool, optional
          use the chisq all times instead of re-calculate 
        
       

        Returns
        -------
        amp : float
            The optimum amplitude calculated for the trace (in Amps)
            with no time shifting allowed (or at the time specified by
            'shift_usec').

         t0 : float
            The time shift (=0 or shift_usec)

        chisq : float
            The chi^2 value calculated from the optimum filter with no
            time shifting (or at the time shift specified by shift_usec)

        """

        # intialize
        amp = np.nan
        chisq = np.nan

        # check pretrigger
        if self._pretrigger_samples  is None:
            self._pretrigger_samples = self._nbins//2
        
        # shift
        t0 = 0
        t0_ind = self._pretrigger_samples
        if shift_usec is not None:
            t0 =  shift_usec*1e-6
            t0_ind += round(t0*self._fs)
            if t0_ind<0:
                t0_ind = 0
            elif t0_ind>self._nbins-1:
                t0_ind = self._nbins-1
            
            

        # use already calculated chisq/amp array
        if use_chisq_alltimes:

            # check if available
            if (not self._chisqs_alltimes_rolled
                or template_tag not in self._chisqs_alltimes_rolled.keys()):
                self.calc_chisq_amp(template_tags=template_tag)
                
            amp = self._amps_alltimes_rolled[template_tag][t0_ind]
            chisq = self._chisqs_alltimes_rolled[template_tag][t0_ind]
    
        else:
            
            # check if filtered signal available
            # and chisq0 available
            if (not self._signal_filts
                or  template_tag not in  self._signal_filts.keys()):
                self.calc_signal_filt(template_tags=template_tag)
                
            if  self._chisq0 is None:
                self.calc_chisq0()

            signal_filt = self._signal_filts[template_tag]
            
            # amplitude
            
            if shift_usec is not None:
                amp = np.real(np.sum(
                    signal_filt*np.exp(2.0j*np.pi*t0*self._fft_freqs),
                    axis=-1,
                ))*self._df
                
            else:
                amp = np.real(np.sum(
                    signal_filt, axis=-1
                ))*self._df
                

            # total chisq
            chisq = self._chisq0 - (amp**2)*self._norms[template_tag]
            
            

        return amp, t0, chisq





    
    def get_fit_withdelay(self, template_tag='default',
                          window_min_from_trig_usec=None,
                          window_max_from_trig_usec=None,
                          window_min_index=None,
                          window_max_index=None,
                          lgc_outside_window=False,
                          pulse_direction_constraint=0,
                          interpolate_t0=False):
        """
        Function for calculating the optimum amplitude of a pulse in
        data with time delay. The OF window min/max can be specified 
        either in usec from pretrigger or ADC samples. If no window, 
        the all trace (unconstrained) is used. 

        Parameters
        ----------

        template_tag : str, optional
          template tag/id
          default: 'default'

               
        window_min_from_trig_usec : float, optional
           OF filter window start in micro seconds from
           pre-trigger (can be negative if prior pre-trigger)


        window_max_from_trig_usec : float, optional
           OF filter window end in micro seconds from
           pre-trigger (can be negative if prior pre-trigger)


        window_min_index: int, optional
            OF filter window start in ADC samples 
        
        window_max_index: int, optional
            OF filter window end in ADC samples

        lgcoutsidewindow : bool, optional
            Boolean flag that is used to specify whether the Optimum
            Filter should look inside `nconstrain` or outside it. If
            False, the filter will minimize the chi^2 in the bins
            specified by `nconstrain`, which is the default behavior.
            If True, then it will minimize the chi^2 in the bins that
            do not contain the constrained window.


        pulse_direction_constraint : int, optional
            Sets a constraint on the direction of the fitted pulse.
            If 0, then no constraint on the pulse direction is set.
            If 1, then a positive pulse constraint is set for all fits.
            If -1, then a negative pulse constraint is set for all
            fits. If any other value, then a ValueError will be raised.
      
        interpolate_t0 : bool, optional
            If True, then a precomputed solution to the parabolic
            equation is used to find the interpolated time-of-best-fit.
            Default is False.

        Returns
        -------
        amp : float
            The optimum amplitude calculated for the trace (in Amps).
        t0 : float
            The time shift calculated for the pulse (in s).
        chi2 : float
            The chi^2 value calculated from the optimum filter.

        """

        # initialize
        amp = np.nan
        chisq = np.nan
        t0 = np.nan

        # check if chisq available -> if not then calculate
        if (not self._chisqs_alltimes_rolled
            or template_tag not in self._chisqs_alltimes_rolled.keys()):
            self.calc_chisq_amp(template_tags=template_tag)


        
        # check pre-trigger
        if self._pretrigger_samples  is None:
            self._pretrigger_samples = self._nbins//2
     
                  
        # get chisq and amp for all times
        chisqs_all = self._chisqs_alltimes_rolled[template_tag]
        amps_all = self._amps_alltimes_rolled[template_tag]
        


        # mask pulse direction
        constraint_mask = None
        if (pulse_direction_constraint==1
            or pulse_direction_constraint==-1):
            constraint_mask=(
                amps_all*pulse_direction_constraint>0
            )

            
        # find index minimum chisq within window
        window_min = None
        if window_min_from_trig_usec is not None:
            window_min = floor(self._pretrigger_samples
                               + window_min_from_trig_usec*self._fs*1e-6)
        elif window_min_index is not None:
            window_min = window_min_index

        if window_min is not None and window_min<0:
            window_min = 0
            
                

        window_max = None
        if window_max_from_trig_usec is not None:
            window_max = ceil(self._pretrigger_samples
                              + window_max_from_trig_usec*self._fs*1e-6)
        elif window_max_index is not None:
            window_max = window_max_index

        if window_max is not None and window_max>self._nbins:
            window_max = self._nbins


        bestind = argmin_chisq(
            chisqs_all,
            window_min=window_min,
            window_max=window_max,
            lgc_outside_window=lgc_outside_window,
            constraint_mask=constraint_mask
        )


        # extract chisq/amp (interpolate if requested)
        if np.isnan(bestind):
            amp = 0.0
            t0 = 0.0
            chisq = self._chisq0
        elif interpolate_t0:
            amp, dt_interp, chisq = interpolate_of(
                amps_all, chisqs_all, bestind, 1/self._fs,
            )
            t0 = (bestind-self._pretrigger_samples)/self._fs + dt_interp
        else:
            amp = amps_all[bestind]
            t0 = (bestind-self._pretrigger_samples)/self._fs
            chisq = chisqs_all[bestind]
           
            
        return amp, t0, chisq


    
    def get_energy_resolution(self,  template_tag='default'):
        """
        Method to return the energy resolution for the optimum filter.
        (resolution depends only on template and noise!) for a 
        specified tag
        
        Parameters
        ----------
        
        template_tag : str, optional
          template tag/id
          default: 'default'


        Returns
        -------
        sigma : float
            The energy resolution of the optimum filter.
        """
        
        if (not self._norms
            or template_tag not in self._norms.keys()):
            self.calc_phi(template_tags=template_tag)
            
        sigma =  1.0 / np.sqrt(self._norms[template_tag])
        
        return sigma



    
    def get_time_resolution(self, amp, template_tag='default'):
        """
        Method to return the time resolution for the optimum filter.
        Resolution depends also on fitted amplitude (-> reset every events)
 
        Parameters
        ----------

        amp : float
          OF fitted amplitude
        
        template_tag : str, optional
          template tag/id
          default: 'default'

        Returns
        -------
        sigma : float
            The time resolution of the optimum filter.

        """

        if template_tag not in self._templates_fft.keys():
            raise ValueError('ERROR: Template wit tag "'
                             + template_tag
                             + '" not available!')


        template_fft = self._templates_fft[template_tag]
        
        sigma = 1.0 / np.sqrt(amp**2 * np.sum(
            (2*np.pi*self._fft_freqs)**2 * np.abs(template_fft)**2 / self._psd
        ) * self._df)
        
        return sigma




    
    def get_chisq_lowfreq(self, amp, t0=0, lowchi2_fcutoff=10000,
                          template_tag='default'):
        """
        Method for calculating the low frequency chi^2 of the optimum
        filter, given some cut off frequency.

        Parameters
        ----------
        amp : float
            The optimum amplitude calculated for the trace (in Amps).
        t0 : float, optional
            The time shift calculated for the pulse (in s).
            default: 0 (np shift)
        lowchi2_fcutoff : float, optional
            The frequency (in Hz) that we should cut off the chi^2 when
            calculating the low frequency chi^2. Default is 10 kHz.

        template_tag : str, optional
          template tag/id
          default: 'default'

        Returns
        -------
        chi2low : float
            The low frequency chi^2 value (cut off at lowchi2_fcutoff) for the
            inputted values.

        """

        # template tag
        if template_tag not in self._templates_fft.keys():
            raise ValueError('ERROR: Template wit tag "'
                             + template_tag
                             + '" not available!')

        # check signal
        if self._signal_fft  is None:
            raise ValueError('ERROR: no signal available!')
            
        template_fft = self._templates_fft[template_tag]
        signal_fft = self._signal_fft 

        # calc chisq
        chi2tot = self._df * np.abs(
            signal_fft - amp * np.exp(-2.0j * np.pi * t0 * self._fft_freqs) * template_fft
        )**2 / self._psd
        

        # find low freq indices
        chi2inds = np.abs(self._fft_freqs) <= lowchi2_fcutoff

        # sum
        chi2low = np.sum(chi2tot[chi2inds])
    
        return chi2low

    


    
   



import numpy as np
import matplotlib.pyplot as plt

from ._base_didv import squarewaveresponse, complexadmittance


class _PlotDIDV(object):
    """Class that contains all plotting functions for DIDV."""

    def _get_best_time_offset(self):
        """
        Helper method for returning the time offset value that
        corresponds to the fit with the lowest chi-square.

        """

        cost_lambda = lambda x: x['cost'] if x is not None else None
        dt_lambda = lambda x: x['params']['dt'] if x is not None else None

        cost_vals = [
            cost_lambda(self._1poleresult),
            cost_lambda(self._2poleresult),
            cost_lambda(self._3poleresult),
        ]
        dt_vals = [
            dt_lambda(self._1poleresult),
            dt_lambda(self._2poleresult),
            dt_lambda(self._3poleresult),
        ]
        if all(fv is None for fv in cost_vals):
            best_time_offset = 0
        else:
            min_cost_idx = min(
                (val, ii) for ii, val in enumerate(
                    cost_vals
                ) if val is not None
            )[1]
            best_time_offset = dt_vals[min_cost_idx]

        return best_time_offset


    def _plot_time_domain(self, poles):
        """Helper function for plotting the fits in time domain."""

        if poles == "all":
            poleslist = np.array([1, 2, 3])
        else:
            poleslist = np.array(poles)

        ## plot the entire trace with fits
        fig, ax = plt.subplots(figsize=(10, 6))

        ax.plot(
            self._time * 1e6,
            (self._tmean - self._offset) * 1e6,
            color='k',
            label='Mean',
        )

        if (self._1poleresult is not None) and (1 in poleslist):
            if 'smallsignalparams' in self._1poleresult:
                key = 'smallsignalparams'
            else:
                key = 'params'
            didvfit1_timedomain = squarewaveresponse(
                self._time,
                self._sgamp,
                self._sgfreq,
                self._dutycycle,
                **self._1poleresult[key],
            )
            ax.plot(
                (self._time + self._1poleresult['params']['dt']) * 1e6,
                didvfit1_timedomain * 1e6,
                color='magenta',
                alpha=0.9,
                label='1-Pole Fit',
            )

        if (self._2poleresult is not None) and (2 in poleslist):
            if 'smallsignalparams' in self._2poleresult:
                key = 'smallsignalparams'
            else:
                key = 'params'
            didvfit2_timedomain = squarewaveresponse(
                self._time,
                self._sgamp,
                self._sgfreq,
                self._dutycycle,
                **self._2poleresult[key],
            )
            ax.plot(
                (self._time + self._2poleresult['params']['dt']) * 1e6,
                didvfit2_timedomain * 1e6,
                color='green',
                alpha=0.9,
                label='2-Pole Fit',
            )

        if (self._3poleresult is not None) and (3 in poleslist):
            if 'smallsignalparams' in self._3poleresult:
                key = 'smallsignalparams'
            else:
                key = 'params'
            didvfit3_timedomain = squarewaveresponse(
                self._time,
                self._sgamp,
                self._sgfreq,
                self._dutycycle,
                **self._3poleresult[key],
            )
            ax.plot(
                (self._time + self._3poleresult['params']['dt']) * 1e6,
                didvfit3_timedomain * 1e6,
                color='orange',
                alpha=0.9,
                label='3-Pole Fit',
            )

        ax.set_xlabel('Time ($\mu$s)')
        ax.set_ylabel('Amplitude ($\mu$A)')
        ax.legend(loc='upper left')
        ax.grid(linestyle='dotted')
        ax.tick_params(which='both', direction='in', right=True, top=True)

        return fig, ax


    def plot_full_trace(self, poles="all", saveplot=False, savepath="",
                        savename=""):
        """
        Function to plot the entire trace in time domain

        Parameters
        ----------
        poles : int, string, array_like, optional
            The pole fits that we want to plot. If set to "all", then
            plots all of the fits. Can also be set to just one of the
            fits. Can be set as an array of different fits, e.g. [1, 2]
        saveplot : boolean, optional
            Boolean value on whether or not the figure should be saved
        savepath : string, optional
            Where the figure should be saved. Saved in the current
            directory by default.
        savename : string, optional
            A string to append to the end of the file name if saving.
            Empty string by default.

        """

        fig, ax = self._plot_time_domain(poles)

        ax.set_xlim([self._time[0] * 1e6, self._time[-1] * 1e6])
        ax.set_title("Full Trace of dIdV")

        if saveplot:
            fig.savefig(savepath + f"full_trace_{savename}.png")
            plt.close(fig)
        else:
            plt.show()


    def plot_single_period_of_trace(self, poles="all", saveplot=False,
                                    savepath="", savename=""):
        """
        Function to plot a single period of the trace in time domain

        Parameters
        ----------
        poles : int, string, array_like, optional
            The pole fits that we want to plot. If set to "all", then
            plots all of the fits. Can also be set to just one of the
            fits. Can be set as an array of different fits, e.g. [1, 2]
        saveplot : boolean, optional
            Boolean value on whether or not the figure should be saved
        savepath : string, optional
            Where the figure should be saved. Saved in the current
            directory by default.
        savename : string, optional
            A string to append to the end of the file name if saving.
            Empty string by default.

        """

        fig, ax = self._plot_time_domain(poles)

        period = 1.0/self._sgfreq

        ax.set_xlim([self._time[0] * 1e6, self._time[0] * 1e6 + period * 1e6])
        ax.set_title("Single Period of Trace")

        if saveplot:
            fig.savefig(savepath + f"trace_one_period_{savename}.png")
            plt.close(fig)
        else:
            plt.show()


    def plot_zoomed_in_trace(self, poles="all", zoomfactor=0.1,
                             saveplot=False, savepath="", savename=""):
        """
        Function to plot a zoomed in portion of the trace in time
        domain. This plot zooms in on the overshoot of the DIDV.

        Parameters
        ----------
        poles : int, string, array_like, optional
            The pole fits that we want to plot. If set to "all", then
            plots all of the fits. Can also be set to just one of the
            fits. Can be set as an array of different fits, e.g. [1, 2]
        zoomfactor : float, optional, optional
            Number between zero and 1 to show different amounts of the
            zoomed in trace.
        saveplot : boolean, optional
            Boolean value on whether or not the figure should be saved
        savepath : string, optional
            Where the figure should be saved. Saved in the current
            directory by default.
        savename : string, optional
            A string to append to the end of the file name if saving.
            Empty string by default.

        """

        period = 1.0 / self._sgfreq

        best_time_offset = self._get_best_time_offset()

        fig, ax = self._plot_time_domain(poles)

        ax.set_xlim(
            (best_time_offset + self._time[0] + (
                0.5 - zoomfactor / 2
            ) * period) * 1e6,
            (best_time_offset + self._time[0] + (
                0.5 + zoomfactor / 2) * period
            ) * 1e6,
        )

        ax.set_title("Zoomed In Portion of Trace")

        if saveplot:
            fig.savefig(savepath + f"zoomed_in_trace_{savename}.png")
            plt.close(fig)
        else:
            plt.show()


    def plot_didv_flipped(self, poles="all", saveplot=False, savepath="",
                          savename=""):
        """
        Function to plot the flipped trace in time domain. This
        function should be used to test if there are nonlinearities in
        the didv.

        Parameters
        ----------
        poles : int, string, array_like, optional
            The pole fits that we want to plot. If set to "all", then
            plots all of the fits. Can also be set to just one of the
            fits. Can be set as an array of different fits, e.g. [1, 2]
        saveplot : boolean, optional
            Boolean value on whether or not the figure should be saved
        savepath : string, optional
            Where the figure should be saved. Saved in the current
            directory by default.
        savename : string, optional
            A string to append to the end of the file name if saving.
            Empty string by default.

        """

        fig, ax = self._plot_time_domain(poles)

        period = 1.0 / self._sgfreq
        time_flipped = self._time - period / 2.0
        tmean_flipped = -(self._tmean - self._offset)

        ax.plot(
            time_flipped * 1e6,
            tmean_flipped * 1e6,
            color='blue',
            label='Flipped Data',
        )

        ax.set_title("Flipped Traces to Check Asymmetry")

        if saveplot:
            fig.savefig(savepath + f"flipped_trace_{savename}.png")
            plt.close(fig)
        else:
            plt.show()


    def _plot_freq_domain(self, function, poles):
        """Helper method for plotting data in frequency domain."""
        
        if poles == "all":
            poleslist = np.array([1, 2, 3])
        else:
            poleslist = np.array(poles)

        # remove values set with placeholder mean
        goodinds = (
            self._didvmean != 0.5 - 0.5j
        ) & (
            self._didvmean != - 0.5 + 0.5j
        )
        fitinds = self._freq > 0
        plotinds = np.logical_and(fitinds, goodinds)

        best_time_offset = self._get_best_time_offset()

        time_phase = np.exp(2.0j * np.pi * best_time_offset * self._freq)

        ## plot the real part of the dIdV in frequency domain
        fig, ax = plt.subplots(figsize=(10, 6))

        ax.scatter(
            self._freq[plotinds],
            function(self._didvmean * time_phase)[plotinds],
            color='blue',
            label='Mean',
            s=5,
        )
        ## plot error in real part of dIdV
        ax.plot(
            self._freq[plotinds],
            function((self._didvmean + self._didvstd) * time_phase)[plotinds],
            color='black',
            label='1-$\sigma$ Bounds',
            alpha=0.1,
        )
        ax.plot(
            self._freq[plotinds],
            function((self._didvmean - self._didvstd) * time_phase)[plotinds],
            color='black',
            alpha=0.1,
        )

        if (self._1poleresult is not None) and (1 in poleslist):
            if 'smallsignalparams' in self._1poleresult:
                key = 'smallsignalparams'
            else:
                key = 'params'
            didvfit1_freqdomain = complexadmittance(
                self._freq, **self._1poleresult[key],
            )
            ax.plot(
                self._freq[fitinds],
                function(didvfit1_freqdomain)[fitinds],
                color='magenta',
                label='1-Pole Fit',
            )

        if (self._2poleresult is not None) and (2 in poleslist):
            if 'smallsignalparams' in self._2poleresult:
                key = 'smallsignalparams'
            else:
                key = 'params'
            didvfit2_freqdomain = complexadmittance(
                self._freq, **self._2poleresult[key],
            )
            ax.plot(
                self._freq[fitinds],
                function(didvfit2_freqdomain)[fitinds],
                color='green',
                label='2-Pole Fit',
            )

        if (self._3poleresult is not None) and (3 in poleslist):
            if 'smallsignalparams' in self._3poleresult:
                key = 'smallsignalparams'
            else:
                key = 'params'
            didvfit3_freqdomain = complexadmittance(
                self._freq, **self._3poleresult[key],
            )
            ax.plot(
                self._freq[fitinds],
                function(didvfit3_freqdomain)[fitinds],
                color='orange',
                label='3-Pole Fit',
            )

        ax.set_xlabel('Frequency (Hz)')
        ax.set_xscale('log')

        yhigh = max(function(
            self._didvmean * time_phase
        )[plotinds][self._freq[plotinds] < 1e5])
        ylow = min(function(
            self._didvmean * time_phase
        )[plotinds][self._freq[plotinds] < 1e5])

        ybnd = np.max([yhigh, -ylow])

        ax.set_ylim([-ybnd, ybnd])
        ax.set_xlim([min(self._freq[fitinds]), max(self._freq[fitinds])])
        ax.legend(loc='upper left')
        ax.tick_params(which='both', direction='in', right=True, top=True)
        ax.grid(which='major')
        ax.grid(which='minor', linestyle='dotted', alpha=0.3)

        return fig, ax


    def plot_re_im_didv(self, poles="all", saveplot=False, savepath="",
                        savename=""):
        """
        Function to plot the real and imaginary parts of the didv in
        frequency space. Currently creates two different plots.

        Parameters
        ----------
        poles : int, string, array_like, optional
            The pole fits that we want to plot. If set to "all", then
            plots all of the fits. Can also be set to just one of the
            fits. Can be set as an array of different fits, e.g. [1, 2]
        saveplot : boolean, optional
            Boolean value on whether or not the figure should be saved
        savepath : string, optional
            Where the figure should be saved. Saved in the current
            directory by default.
        savename : string, optional
            A string to append to the end of the file name if saving.
            Empty string by default.

        """

        fig, ax = self._plot_freq_domain(np.real, poles)

        ax.set_title("Real Part of dIdV")
        ax.set_ylabel('Re($dI/dV$) ($\Omega^{-1}$)')

        if saveplot:
            fig.savefig(savepath + f"didv_real_{savename}.png")
            plt.close(fig)
        else:
            plt.show()

        
        fig, ax = self._plot_freq_domain(np.imag, poles)

        ax.set_title("Imaginary Part of dIdV")
        ax.set_ylabel('Im($dI/dV$) ($\Omega^{-1}$)')

        if saveplot:
            fig.savefig(savepath + f"didv_imag_{savename}.png")
            plt.close(fig)
        else:
            plt.show()


    def plot_abs_phase_didv(self, poles="all", saveplot=False, savepath="",
                            savename=""):
        """
        Function to plot the absolute value and the phase of the dIdV
        in frequency space. Currently creates two different plots.

        Parameters
        ----------
        poles : int, string, array_like, optional
            The pole fits that we want to plot. If set to "all", then
            plots all of the fits. Can also be set to just one of the
            fits. Can be set as an array of different fits, e.g. [1, 2]
        saveplot : boolean, optional
            Boolean value on whether or not the figure should be saved
        savepath : string, optional
            Where the figure should be saved. Saved in the current
            directory by default.
        savename : string, optional
            A string to append to the end of the file name if saving.
            Empty string by default.

        """

        fig, ax = self._plot_freq_domain(np.abs, poles)

        ax.set_title("|dIdV|")
        ax.set_ylabel('Abs($dI/dV$) ($\Omega^{-1}$)')
        ax.set_ylim(0)

        if saveplot:
            fig.savefig(savepath + f"didv_abs_{savename}.png")
            plt.close(fig)
        else:
            plt.show()


        fig, ax = self._plot_freq_domain(np.angle, poles)

        ax.set_title("Phase of dIdV")
        ax.set_ylabel('Arg($dI/dV$)')
        ax.set_ylim(-np.pi, np.pi)

        if saveplot:
            fig.savefig(savepath + f"didv_phase_{savename}.png")
            plt.close(fig)
        else:
            plt.show()


    def plot_re_vs_im_dvdi(self, poles='all', saveplot=False, savepath="",
                           savename=""):
        """
        Function to plot the real vs imaginary parts of the complex
        impedance.

        Parameters
        ----------
        poles : int, string, array_like, optional
            The pole fits that we want to plot. If set to "all", then
            plots all of the fits. Can also be set to just one of the
            fits. Can be set as an array of different fits, e.g. [1, 2]
        saveplot : boolean, optional
            Boolean value on whether or not the figure should be saved
        savepath : string, optional
            Where the figure should be saved. Saved in the current
            directory by default.
        savename : string, optional
            A string to append to the end of the file name if saving.
            Empty string by default.

        """
        
        if poles == "all":
            poleslist = np.array([1, 2, 3])
        else:
            poleslist = np.array(poles)

        ## don't plot points with huge errors
        goodinds = np.abs(self._didvmean / self._didvstd) > 2.0
        fitinds = self._freq > 0
        plotinds = np.logical_and(fitinds, goodinds)

        best_time_offset = self._get_best_time_offset()

        time_phase = np.exp(2.0j * np.pi * best_time_offset * self._freq)

        fig, ax = plt.subplots(figsize=(10, 6))

        ax.scatter(
            np.real(1 / (self._didvmean * time_phase))[plotinds],
            np.imag(1 / (self._didvmean * time_phase))[plotinds],
            color='blue',
            label='Mean',
            s=5,
        )

        if (self._1poleresult is not None) and (1 in poleslist):
            if 'smallsignalparams' in self._1poleresult:
                key = 'smallsignalparams'
            else:
                key = 'params'
            didvfit1_freqdomain = complexadmittance(
                self._freq, **self._1poleresult[key],
            )
            ax.plot(
                np.real(1 / didvfit1_freqdomain)[fitinds],
                np.imag(1 / didvfit1_freqdomain)[fitinds],
                color='magenta',
                label='1-Pole Fit',
            )

        if (self._2poleresult is not None) and (2 in poleslist):
            if 'smallsignalparams' in self._2poleresult:
                key = 'smallsignalparams'
            else:
                key = 'params'
            didvfit2_freqdomain = complexadmittance(
                self._freq, **self._2poleresult[key],
            )
            ax.plot(
                np.real(1 / didvfit2_freqdomain)[fitinds],
                np.imag(1 / didvfit2_freqdomain)[fitinds],
                color='green',
                label='2-Pole Fit',
            )

        if (self._3poleresult is not None) and (3 in poleslist):
            if 'smallsignalparams' in self._3poleresult:
                key = 'smallsignalparams'
            else:
                key = 'params'
            didvfit3_freqdomain = complexadmittance(
                self._freq, **self._3poleresult[key],
            )
            ax.plot(
                np.real(1 / didvfit3_freqdomain)[fitinds],
                np.imag(1 / didvfit3_freqdomain)[fitinds],
                color='orange',
                label='3-Pole Fit',
            )

        ax.set_xlabel('Re($dV/dI$) ($\Omega$)')
        ax.set_ylabel('Im($dV/dI$) ($\Omega$)')
        ax.set_title('Re($dV/dI$) vs. Im($dV/dI$)')

        ax.legend(loc='upper left')
        ax.tick_params(which='both', direction='in', right=True, top=True)
        ax.grid(which='major')
        ax.grid(which='minor', linestyle='dotted', alpha=0.3)

        if saveplot:
            fig.savefig(savepath+f"dvdi_real_vs_imag_{savename}.png")
            plt.close(fig)
        else:
            plt.show()



import numpy as np
from math import ceil
from scipy.signal import savgol_filter
import matplotlib.pyplot as plt


__all__ = [
    "plot_psd",
    "plot_reim_psd",
    "plot_corrcoeff",
    "plot_csd",
    "plot_decorrelatednoise",
    "compare_noise",
    "plot_noise_sim",
]


def plot_psd(noise, lgcoverlay=True, lgcsave=False, savepath=None):
    """
    Function to plot the noise spectrum referenced to the TES line in units of Amperes/sqrt(Hz).

    Parameters
    ----------
    noise : Object
        Noise object to be plotted
    lgcoverlay : boolean, optional
        If True, psd's for all channels are overlayed in a single plot,
        If False, each psd for each channel is plotted in a seperate subplot
    lgcsave : boolean, optional
        If True, the figure is saved in the user provided directory
    savepath : str, optional
        Absolute path for the figure to be saved

    """

    if noise.psd is None:
        print('Need to calculate the psd first')
        return
    else:
        ### Overlay plot
        if lgcoverlay:
            plt.figure(figsize=(12, 8))
            plt.title(f'{noise.name} PSD')
            plt.xlabel('Frequency [Hz]')
            plt.ylabel(r'Input Referenced Noise [A/$\sqrt{\mathrm{Hz}}$]')
            plt.grid(which='both')
            for ichan, channel in enumerate(noise.channames):
                plt.loglog(noise.freqs[1:], np.sqrt(noise.psd[ichan][1:]), label=channel)
            lgd = plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0)
            if lgcsave:
                try:
                    plt.savefig(
                        savepath + noise.name.replace(" ", "_") + '_PSD_overlay.png',
                        bbox_extra_artists=(lgd, ),
                        bbox_inches='tight',
                    )
                except:
                    print('Invalid save path. Figure not saved')
            plt.show()
        ### Subplots            
        else:
            num_subplots = len(noise.channames)
            nrows = int(ceil(num_subplots / 2))
            ncolumns = 2
            fig, axes = plt.subplots(nrows, ncolumns, figsize=(6 * num_subplots, 6 * num_subplots))
            plt.suptitle(f'{noise.name} PSD', fontsize=40)
            for ii in range(nrows * 2):
                if ii < nrows:
                    irow = ii
                    jcolumn = 0
                else:
                    irow = ii - nrows
                    jcolumn = 1
                if ii < num_subplots and nrows > 1:    
                    axes[irow,jcolumn].set_title(noise.channames[ii], fontsize=30)
                    axes[irow,jcolumn].set_xlabel('Frequency [Hz]', fontsize=25)
                    axes[irow,jcolumn].set_ylabel(r'Input Referenced Noise [A/$\sqrt{\mathrm{Hz}}$]', fontsize=25)
                    axes[irow,jcolumn].grid(which='both')
                    axes[irow,jcolumn].loglog(noise.freqs[1:], np.sqrt(noise.psd[ii][1:]))
                elif ii < num_subplots and nrows==1:
                    axes[jcolumn].set_title(noise.channames[ii], fontsize=30)
                    axes[jcolumn].set_xlabel('Frequency [Hz]', fontsize=25)
                    axes[jcolumn].set_ylabel(r'Input Referenced Noise [A/$\sqrt{\mathrm{Hz}}$]', fontsize=25)
                    axes[jcolumn].grid(which = 'both')
                    axes[jcolumn].loglog(noise.freqs[1:], np.sqrt(noise.psd[ii][1:]))
                elif nrows==1:
                    axes[jcolumn].axis('off')
                else:
                    axes[irow,jcolumn].axis('off')
            plt.tight_layout() 
            plt.subplots_adjust(top=0.95)

            if lgcsave:
                try:
                    plt.savefig(savepath + noise.name.replace(" ", "_") + '_PSD_subplot.png')
                except:
                    print('Invalid save path. Figure not saved')
            plt.show()



def plot_reim_psd(noise, lgcsave=False, savepath=None):
    """
    Function to plot the real vs imaginary noise spectrum referenced to the TES line in units of Amperes/sqrt(Hz).
    This is done to check for thermal muon tails making it passed the quality cuts

    Parameters
    ----------
    noise : Object
        Noise object to be plotted
    lgcsave : boolean, optional
        If True, the figure is saved in the user provided directory
    savepath : str, optional
        Absolute path for the figure to be saved

    """

    if noise.real_psd is None:
        print('Need to calculate the psd first')
        return
    else:
        num_subplots = len(noise.channames)
        nrows = int(ceil(num_subplots / 2))
        ncolumns = 2
        fig, axes = plt.subplots(nrows, ncolumns, figsize=(6 * num_subplots, 6 * num_subplots)) 
        plt.suptitle('{} Real vs Imaginary PSD'.format(noise.name), fontsize=40)
        for ii in range(nrows * 2):
            if ii < nrows:
                irow = ii
                jcolumn = 0
            else:
                irow = ii - nrows
                jcolumn = 1
            if ii < num_subplots and nrows > 1:
                axes[irow,jcolumn].set_title(noise.channames[ii], fontsize=30)
                axes[irow,jcolumn].set_xlabel('Frequency [Hz]', fontsize=25)
                axes[irow,jcolumn].set_ylabel(r'Input Referenced Noise [A/$\sqrt{\mathrm{Hz}}$]', fontsize=25)
                axes[irow,jcolumn].grid(which='both')
                axes[irow,jcolumn].loglog(noise.freqs[1:], np.sqrt(noise.real_psd[ii][1:]), label='real')
                axes[irow,jcolumn].loglog(noise.freqs[1:], np.sqrt(noise.imag_psd[ii][1:]), label='imag')
                axes[irow,jcolumn].legend()
            elif ii < num_subplots and nrows==1:
                axes[jcolumn].set_title(noise.channames[ii], fontsize=30)
                axes[jcolumn].set_xlabel('Frequency [Hz]', fontsize=25)
                axes[jcolumn].set_ylabel(r'Input Referenced Noise [A/$\sqrt{\mathrm{Hz}}$]', fontsize=25)
                axes[jcolumn].grid(which='both')
                axes[jcolumn].loglog(noise.freqs[1:], np.sqrt(noise.real_psd[ii][1:]), label='real')
                axes[jcolumn].loglog(noise.freqs[1:], np.sqrt(noise.imag_psd[ii][1:]), label='imag')
                axes[jcolumn].legend()
            elif nrows==1:
                axes[jcolumn].axis('off')
            else:
                axes[irow,jcolumn].axis('off')
        plt.tight_layout() 
        plt.subplots_adjust(top=0.95)

        if lgcsave:
            try:
                plt.savefig(savepath + noise.name.replace(" ", "_") + '_ReIm_PSD.png')
            except:
                print('Invalid save path. Figure not saved')
        plt.show()        



def plot_corrcoeff(noise, lgcsmooth=True, nwindow=7, lgcsave=False, savepath=None):
    """
    Function to plot the cross channel correlation coefficients. Since there are typically few traces,
    the correlations are often noisy. a savgol_filter is used to smooth out some of the noise

    Parameters
    ----------
    noise : Object
        noise object to be plotted
    lgcsmooth : boolean, optional
        If True, a savgol_filter will be used when plotting.
    nwindow : int, optional
        the number of bins used for the window in the savgol_filter
    lgcsave : boolean, optional
        If True, the figure is saved in the user provided directory
    savepath : str, optional
        Absolute path for the figure to be saved

    """

    if (noise.corrcoeff is None):
        print('Need to calculate the corrcoeff first')
        return
    else:
        plt.figure(figsize=(12, 8))
        plt.title('{} \n Cross Channel Correlation Coefficients'.format(noise.name))
        for ii in range(noise.corrcoeff.shape[0]):
            for jj in range(noise.corrcoeff.shape[1]):
                if ii > jj:
                    label = '{} - {}'.format(noise.channames[ii], noise.channames[jj])
                    if lgcsmooth:
                        plt.plot(
                            noise.freqs[1:],
                            savgol_filter(noise.corrcoeff[ii][jj][1:], nwindow, 3, mode = 'nearest'),
                            label=label,
                            alpha=0.5,
                        )
                    else:
                        plt.plot(noise.freqs[1:], noise.corrcoeff[ii][jj][1:] , label=label, alpha=0.5)
                        
                    plt.xscale('log')
        plt.xlabel('frequency [Hz]')
        plt.ylabel(r'Correlation Coeff [COV(x,y)/$\sigma_x \sigma_y$]')
        plt.grid(which = 'both')
        lgd = plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=1.)

        if lgcsave:
            try:
                plt.savefig(
                    savepath + noise.name.replace(" ", "_") + '_corrcoeff.png',
                    bbox_extra_artists=(lgd, ),
                    bbox_inches='tight',
                )
            except:
                print('Invalid save path. Figure not saved')

        plt.show()

def plot_csd(noise, whichcsd=['01'], lgcreal=True, lgcsave=False, savepath=None):
    """
    Function to plot the cross channel noise spectrum referenced to the TES line in
    units of Amperes^2/Hz

    Parameters
    ----------
    noise : Object
        Noise object to be plotted
    whichcsd : list, optional
        a list of strings, where each element of the list refers to the pair of 
        indices of the desired csd plot
    lgcreal : boolean, optional
        If True, the Re(csd) is plotted. If False, the Im(csd) is plotted
    lgcsave : boolean, optional
        If True, the figure is saved in the user provided directory
    savepath : str, optional
        Absolute path for the figure to be saved

    """

    if noise.csd is None:
        print('Must calculate the csd first')
        return
    else:
        x_plt_label = []
        y_plt_label = []
        for label in whichcsd:
            if type(label) == str:
                x_plt_label.append(int(label[0]))
                y_plt_label.append(int(label[1]))
                if ((int(label[0]) > noise.real_csd.shape[0] - 1) or (int(label[1]) > noise.real_csd.shape[1] - 1)):
                    print('index out of range')
                    return
            else:
                print("Invalid selection. Please provide a list of strings for the desired plots. Ex: ['01','02'] ")
                return

        for ii in range(len(x_plt_label)):
            plt.figure(figsize=(12, 8))
            if lgcreal:
                title = '{} Re(CSD) for channels: {}-{}'.format(
                    noise.name,
                    noise.channames[x_plt_label[ii]],
                    noise.channames[y_plt_label[ii]],
                )

                plt.loglog(noise.freqs[1:], noise.real_csd[x_plt_label[ii]][y_plt_label[ii]][1:])
            else:
                title = '{} Im(CSD) for channels: {}-{}'.format(
                    noise.name,
                    noise.channames[x_plt_label[ii]],
                    noise.channames[y_plt_label[ii]],
                )

                plt.loglog(noise.freqs[1:], noise.imag_csd[x_plt_label[ii]][y_plt_label[ii]][1:])
            plt.title(title)
            plt.grid(which='both')
            plt.xlabel('Frequency [Hz]')
            plt.ylabel(r'CSD [A$^2$/Hz]')

            if lgcsave:
                try:
                    plt.savefig(savepath + noise.name.replace(" ", "_") + '_csd{}.png'.format(ii))
                except:
                    print('Invalid save path. Figure not saved')
            plt.show()


def plot_decorrelatednoise(noise, lgcoverlay=False, lgcdata=True, lgcuncorrnoise=True, lgccorrelated=False,
                           lgcsum=False, lgcsave=False, savepath=None):
    """
    Function to plot the de-correlated noise spectrum referenced to the TES line in units of Amperes/sqrt(Hz) 
    from fitted parameters calculated calculate_deCorrelated_noise

    Parameters
    ----------
    noise : Object
        Noise object to be plotted
    lgcoverlay : boolean, optional
        If True, de-correlated for all channels are overlayed in a single plot,
        If False, the noise for each channel is plotted in a seperate subplot
    lgcdata : boolean, optional
        Only applies when lgcoverlay = False. If True, the csd data is plotted
    lgcuncorrnoise : boolean, optional
        Only applies when lgcoverlay = False. If True, the de-correlated noise is plotted
    lgccorrelated : boolean, optional
        Only applies when lgcoverlay = False. If True, the correlated component of the fitted noise
        is plotted
    lgcsum : boolean, optional
        Only applies when lgcoverlay = False. If True, the sum of the fitted de-correlated noise and
        and correlated noise is plotted
    lgcsave : boolean, optional
        If True, the figure is saved in the user provided directory
    savepath : str, optional
        Absolute path for the figure to be saved

    """

    if noise.uncorrnoise is None:
        print('Need to de-correlate the noise first')
        return
    else:

        ### Overlay plot
        if lgcoverlay:
            plt.figure(figsize=(12, 8))
            plt.xlabel('frequency [Hz]')
            plt.ylabel(r'Input Referenced Noise [A/$\sqrt{\mathrm{Hz}}$]')
            plt.grid(which='both')
            plt.title('{} de-correlated noise'.format(noise.name))
            for ichan, channel in enumerate(noise.channames):
                    plt.loglog(noise.freqs[1:], np.sqrt(noise.uncorrnoise[ichan][1:]), label=channel)
            lgd = plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0)

            if lgcsave:
                try:
                    plt.savefig(
                        savepath + noise.name.replace(" ", "_") + '_deCorrNoise_overlay.png',
                        bbox_extra_artists=(lgd, ),
                        bbox_inches='tight',
                    ) 
                except:
                    print('Invalid save path. Figure not saved')
            plt.show()
        ### Subplots
        else:
            num_subplots = len(noise.channames)
            nrows = int(ceil(num_subplots / 2))
            ncolumns = 2
            fig, axes = plt.subplots(nrows, ncolumns, figsize=(6 * num_subplots, 6 * num_subplots)) 
            plt.suptitle('{} de-correlated noise'.format(noise.name), fontsize=40)
            for ii in range(nrows * 2):
                if ii < nrows:
                    irow = ii
                    jcolumn = 0
                else:
                    irow = ii - nrows
                    jcolumn = 1
                if ii < num_subplots and nrows > 1:    
                    axes[irow,jcolumn].set_title(noise.channames[ii], fontsize=30)
                    axes[irow,jcolumn].set_xlabel('Frequency [Hz]', fontsize=25)
                    axes[irow,jcolumn].set_ylabel(r'Input Referenced Noise [A/$\sqrt{\mathrm{Hz}}$]', fontsize=25)
                    axes[irow,jcolumn].grid(which='both')
                    if lgcdata:
                        axes[irow,jcolumn].loglog(
                            noise.freqs[1:],
                            np.sqrt(noise.real_csd[ii][ii][1:]),
                            label='Data',
                            alpha=0.4,
                        )
                    if lgcuncorrnoise:
                        axes[irow,jcolumn].loglog(
                            noise.freqs[1:],
                            np.sqrt(noise.uncorrnoise[ii][1:]),
                            label='Uncorrelated Noise',alpha = 0.6)
                    if lgccorrelated:
                        axes[irow,jcolumn].loglog(
                            noise.freqs[1:],
                            np.sqrt(noise.corrnoise[ii][1:]),
                            label='Correlated Noise',
                            alpha=0.6,
                        )
                    if lgcsum:
                        axes[irow,jcolumn].loglog(
                            noise.freqs[1:],
                            np.sqrt(noise.uncorrnoise[ii][1:] + noise.corrnoise[ii][1:]),
                            label='Total Noise',
                            alpha=0.6,
                        )
                    axes[irow,jcolumn].legend()
                elif ii < num_subplots and nrows==1:
                    axes[jcolumn].set_title(noise.channames[ii], fontsize=30)
                    axes[jcolumn].set_xlabel('Frequency [Hz]', fontsize=25)
                    axes[jcolumn].set_ylabel(r'Input Referenced Noise [A/$\sqrt{\mathrm{Hz}}$]', fontsize=25)
                    axes[jcolumn].grid(which='both')
                    if lgcdata:
                        axes[jcolumn].loglog(
                            noise.freqs[1:],
                            np.sqrt(noise.real_csd[ii][ii][1:]),
                            label='Data',
                            alpha=0.4,
                        )
                    if lgcuncorrnoise:
                        axes[jcolumn].loglog(
                            noise.freqs[1:],
                            np.sqrt(noise.uncorrnoise[ii][1:]),
                            label='Uncorrelated Noise',
                            alpha=0.6,
                        )
                    if lgccorrelated:
                        axes[jcolumn].loglog(
                            noise.freqs[1:],
                            np.sqrt(noise.corrnoise[ii][1:]),
                            label='Correlated Noise',
                            alpha=0.6,
                        )
                    if lgcsum:
                        axes[jcolumn].loglog(
                            noise.freqs[1:],
                            np.sqrt(noise.uncorrnoise[ii][1:] + noise.corrnoise[ii][1:]),
                            label='Total Noise',
                            alpha=0.6,
                        )
                    axes[jcolumn].legend()
                elif nrows==1:
                    axes[jcolumn].axis('off')
                else:
                    axes[irow,jcolumn].axis('off')
            plt.tight_layout() 
            plt.subplots_adjust(top=0.95)

            if lgcsave:
                try:
                    plt.savefig(savepath + noise.name.replace(" ", "_") + '_deCorrNoise_subplot.png')
                except:
                    print('Invalid save path. Figure not saved')
            plt.show()


def compare_noise(arr, channels, lgcdecorrelatednoise=False, lgcsave=False, savepath=None):
    """
    Function to plot multiple psds from different noise objects on the same figure. Each channel will
    be plotted in its own figure.

    Parameters
    ----------
    arr : array_like
        Array of noise objects
    channels : list
        List of strings, each string is a channel to plot. ex ['PSA1','PAS2']
    lgcdecorrelatednoise : boolean, optional
        If False, the psd is for each channel is plotted. If True, the calculated
        de-correlated noise is plotted
    lgcsave : boolean , optional
        If True, the figure is saved in the user provided directory
    savepath : str, optional
        Absolute path for the figure to be saved

    """


    #check to make sure channels is a list or array
    if (type(channels) == np.ndarray or type(channels) == list):
        for chan in channels:
            if chan in set(sum([arr[ii].channames for ii in range(len(arr))], [])): #check if channel is in 
                # any of the noise objects before creating a figure
                plt.figure(figsize=(12, 8))
                plt.title(chan) 
                plt.xlabel('Frequency [Hz]')
                plt.ylabel(r'Input Referenced Noise [A/$\sqrt{\mathrm{Hz}}$]')
                plt.grid(which = 'both')
                for ii in range(len(arr)):
                    if chan in arr[ii].chann_dict:
                        chan_index = arr[ii].chann_dict[chan] #check that the channel is in the noise object before plotting
                        #plot the de correlated noise if desired
                        if lgcdecorrelatednoise:
                            #check that de correlated noise has been calculated
                            if arr[ii].uncorrnoise is not None:
                                plt.loglog(
                                    arr[ii].freqs[1:],
                                    np.sqrt(arr[ii].uncorrnoise[chan_index][1:]),
                                    label=arr[ii].name + ' Decorrelated Noise',
                                )
                            else:
                                print('The decorrelated noise for file: {} has not been calculated yet'.format(arr[ii].name))
                        else:
                            plt.loglog(arr[ii].freqs[1:], np.sqrt(arr[ii].psd[chan_index][1:]), label=arr[ii].name)
                    else:
                        print('channel: {} not found for file: {} '.format(chan, arr[ii].name)) 
                lgd = plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0)
                if lgcsave:
                    try:
                        plt.savefig(
                            savepath + chan + '_PSD_comparison.png',
                            bbox_extra_artists=(lgd, ),
                            bbox_inches='tight',
                        )
                    except:
                        print('Invalid save path. Figure not saved')
                plt.show()
            else:
                print('Invalid channel name: {} '.format(chan))
    else:
        print("Please provide desired channels in format of a list of numpy.ndarray. ex ['PSA1','PAS2']")   





def plot_noise_sim(f, psd, noise_sim, istype, qetbias, lgcsave=False, figsavepath='', xlims=None, ylims=None):
    """
    Plots psd with simulated noise model

    Parameters
    ----------
    f : array_like
        Frequency bins for psd
    psd : array_like 
        Power spectral density
    istype : str
        Must be 'current', 'power', 'sc', or 'normal'
        If 'current' the noise is plotted referenced to TES current
        If 'power' the noise is plotted referenced to TES power
    qetbias : float
        Applied QET bias
    lgcsave : boolean, optional
        If True, plot is saved
    figsavepath : str, optional
        Directory to save trace
    xlims : NoneType, tuple, optional
        Limits to be passed to ax.set_xlim()
    ylims : NoneType, tuple, optional
        Limits to be passed to ax.set_ylim()

    Returns
    -------
    fig : Object
        fig object from matplotlib.pyplot
    ax : Object
        ax object from matplotlib.pyplot

    """

    if istype == 'current':
        fig, ax = _plot_ti_noise(f, psd, noise_sim, xlims, ylims)
        if lgcsave:
            fig.savefig(figsavepath+f'current_qetbias{qetbias*1e6:.3f}muA.png', bbox_inches='tight')
        else:
            return fig, ax 
    elif istype == 'power':
        fig, ax = _plot_tp_noise(f, psd, noise_sim, xlims, ylims)
        if lgcsave:
            fig.savefig(figsavepath+f'power_noise_qetbias{qetbias*1e6:.3f}muA.png', bbox_inches='tight')
        else:
            return fig, ax 
    elif istype == 'sc':
        fig, ax = _plot_sc_noise(f, psd, noise_sim, qetbias, xlims, ylims)    
        if lgcsave:
            plt.savefig(f'{figsavepath}SC_noise_qetbias{qetbias*1e6:.3f}muA.png')
        else:
            return fig, ax 
    elif istype == 'normal':
        fig, ax = _plot_n_noise(f, psd, noise_sim, qetbias, xlims, ylims)    
        if lgcsave:
            plt.savefig(f'{figsavepath}normal_noise_qetbias{qetbias*1e6:.3f}muA.png')
        else:
            return fig, ax 

def _plot_ti_noise(f, psd, noise_sim, xlims, ylims):
    """
    Helper function to plot transition noise in units of current

    Parameters
    ----------
    f : ndarray
        Array of frequency values
    psd : ndarray
        One sided Power spectral density
    noise_sim : TESnoise object
        The noise simulation object
    xlims : NoneType, tuple
        Limits to be passed to ax.set_xlim()
    ylims : NoneType, tuple
        Limits to be passed to ax.set_ylim()

    Returns
    -------
    fig : Object
        fig object from matplotlib.pyplot
    ax : Object
        ax object from matplotlib.pyplot

    """

    freqs = f[1:]
    psd = psd[1:]
    noise_sim.freqs = freqs

    fig, ax = plt.subplots(figsize=(11, 6))
    if xlims is not None:
        ax.set_xlim(xlims)
    if ylims is not None:
        ax.set_ylim(ylims)

    ax.grid(which="major", linestyle='--')
    ax.grid(which="minor", linestyle="dotted", alpha=0.5)
    ax.tick_params(which="both", direction="in", right=True, top=True)
    ax.set_xlabel(r'Frequency [Hz]')
    ax.set_title(f"Current Noise For $R_0$ : {noise_sim.r0*1e3:.2f} $m\Omega$")

    ax.loglog(
        noise_sim.freqs,
        np.sqrt(np.abs(noise_sim.s_ites())),
        color='#1f77b4',
        linewidth=1.5,
        label='TES Johnson Noise',
    )
    ax.loglog(
        noise_sim.freqs,
        np.sqrt(np.abs(noise_sim.s_iload())),
        color='#ff7f0e',
        linewidth=1.5,
        label='Load Noise',
    )
    ax.loglog(
        noise_sim.freqs,
        np.sqrt(np.abs(noise_sim.s_itfn())),
        color='#2ca02c',
        linewidth=1.5,
        label='TFN Noise',
    )
    ax.loglog(
        noise_sim.freqs,
        np.sqrt(np.abs(noise_sim.s_itot())),
        color='#d62728',
        linewidth=1.5,
        label='Total Noise',
    )
    ax.loglog(
        noise_sim.freqs,
        np.sqrt(np.abs(noise_sim.s_isquid())),
        color='#9467bd',
        linewidth=1.5,
        label='Squid+Electronics Noise',
    )
    ax.loglog(freqs, np.sqrt(psd), color='#8c564b', alpha=0.8, label='Raw Data')

    ax.set_ylabel('TES Current Noise $[A/\sqrt{\mathrm{Hz}}]$')

    lgd = plt.legend(loc='upper right')

    return fig, ax
    
    
def _plot_tp_noise(f, psd, noise_sim, xlims, ylims):
    """
    Helper function to plot transition noise in units of power

    Parameters
    ----------
    f : ndarray
        Array of frequency values
    psd : ndarray
        One sided Power spectral density
    noise_sim : TESnoise object
        The noise simulation object
    qetbias : float
        Applied QET bias
    xlims : NoneType, tuple
        Limits to be passed to ax.set_xlim()
    ylims : NoneType, tuple
        Limits to be passed to ax.set_ylim()

    Returns
    -------
    fig : Object
        fig object from matplotlib.pyplot
    ax : Object
        ax object from matplotlib.pyplot

    """

    freqs = f[1:]
    psd = psd[1:]
    noise_sim.freqs = freqs

    fig, ax = plt.subplots(figsize=(11, 6))
    if xlims is not None:
        ax.set_xlim(xlims)
    if ylims is not None:
        ax.set_ylim(ylims)

    ax.grid(which="major", linestyle='--')
    ax.grid(which="minor", linestyle="dotted", alpha=0.5)
    ax.tick_params(which="both", direction="in", right=True, top=True)
    ax.set_xlabel(r'Frequency [Hz]')
    ax.set_title(f"Power Noise For $R_0$ : {noise_sim.r0*1e3:.2f} $m\Omega$")

    ax.loglog(
        noise_sim.freqs,
        np.sqrt(np.abs(noise_sim.s_ptes())),
        color='#1f77b4',
        linewidth=1.5,
        label='TES Johnson Noise',
    )
    ax.loglog(
        noise_sim.freqs,
        np.sqrt(np.abs(noise_sim.s_pload())),
        color='#ff7f0e',
        linewidth=1.5,
        label='Load Noise',
    )
    ax.loglog(
        noise_sim.freqs,
        np.sqrt(np.abs(noise_sim.s_ptfn())),
        color='#2ca02c',
        linewidth=1.5,
        label='TFN Noise',
    )
    ax.loglog(
        noise_sim.freqs,
        np.sqrt(np.abs(noise_sim.s_ptot())),
        color='#d62728',
        linewidth=1.5,
        label='Total Noise',
    )
    ax.loglog(
        noise_sim.freqs,
        np.sqrt(np.abs(noise_sim.s_psquid())),
        color='#9467bd',
        linewidth=1.5,
        label='Squid+Electronics Noise',
    )
    ax.loglog(
        freqs,
        np.sqrt(psd / (np.abs(noise_sim.dIdP(freqs))**2)),
        color='#8c564b',
        alpha=0.8,
        label='Raw Data',
    )
    ax.set_ylabel(r'Input Referenced Power Noise [W/$\sqrt{\mathrm{Hz}}$]')

    lgd = plt.legend(loc='upper right')

    return fig, ax

def _plot_sc_noise(f, psd, noise_sim, qetbias, xlims, ylims):
    """
    Helper function to plot SC noise 

    Parameters
    ----------
    f : ndarray
        Array of frequency values
    psd : ndarray
        One sided Power spectral density
    noise_sim : TESnoise object
        The noise simulation object
    qetbias : float
        Applied QET bias
    xlims : NoneType, tuple
        Limits to be passed to ax.set_xlim()
    ylims : NoneType, tuple
        Limits to be passed to ax.set_ylim()

    Returns
    -------
    fig : Object
        fig object from matplotlib.pyplot
    ax : Object
        ax object from matplotlib.pyplot

    """

    f = f[1:]
    psd = psd[1:]
    fig, ax = plt.subplots(figsize=(11, 6))
    if xlims is not None:
        ax.set_xlim(xlims)
    if ylims is not None:
        ax.set_ylim(ylims)

    ax.grid(which="major", linestyle = '--')
    ax.grid(which="minor", linestyle="dotted", alpha=0.5)
    ax.loglog(
        f, np.sqrt(psd), alpha=0.5, color='#8c564b', label='Raw Data',
    )
    ax.loglog(
        f, np.sqrt(noise_sim.s_isquid(f)), color='#9467bd', label='Squid+Electronics Noise',
    )
    ax.loglog(
        f, np.sqrt(noise_sim.s_iloadsc(f)), color='#ff7f0e', label='Load Noise',
    )
    ax.loglog(
        f, np.sqrt(noise_sim.s_itotsc(f)), color='#d62728', label='Total Noise',
    )
    ax.legend()
    ax.set_xlabel('Frequency [Hz]')
    ax.set_ylabel('Input Referenced Current Noise [A/$\sqrt{\mathrm{Hz}}$]')
    ax.set_title(f'Superconducting State noise for QETbias: {qetbias*1e6} $\mu$A')
    ax.tick_params(which="both", direction="in", right=True, top=True)

    return fig, ax    

def _plot_n_noise(f, psd, noise_sim, qetbias, xlims, ylims):
    """
    Helper function to plot normal state noise 

    Parameters
    ----------
    f : ndarray
        Array of frequency values
    psd : ndarray
        One sided Power spectral density
    noise_sim : TESnoise object
        The noise simulation object
    qetbias : float
        Applied QET bias
    xlims : NoneType, tuple
        Limits to be passed to ax.set_xlim()
    ylims : NoneType, tuple
        Limits to be passed to ax.set_ylim()    

    Returns
    -------
    fig : Object
        fig object from matplotlib.pyplot
    ax : Object
        ax object from matplotlib.pyplot

    """

    f = f[1:]
    psd = psd[1:]
    fig, ax = plt.subplots(figsize=(11, 6))
    if xlims is not None:
        ax.set_xlim(xlims)
    if ylims is not None:
        ax.set_ylim(ylims)

    ax.grid(which="major", linestyle = '--')
    ax.grid(which="minor", linestyle="dotted", alpha=0.5)
    ax.loglog(
        f, np.sqrt(psd), alpha=0.5, color='#8c564b', label='Raw Data',
    )
    ax.loglog(
        f, np.sqrt(noise_sim.s_isquid(f)), color='#9467bd', label='Squid+Electronics Noise',
    )
    ax.loglog(
        f, np.sqrt(noise_sim.s_itesnormal(f)), color='#1f77b4', label= 'TES Johnson Noise',
    )
    ax.loglog(
        f, np.sqrt(noise_sim.s_iloadnormal(f)), color='#ff7f0e', label='Load Noise',
    )
    ax.loglog(
        f, np.sqrt(noise_sim.s_itotnormal(f)), color='#d62728', label='Total Noise',
    )
    ax.legend()
    ax.set_xlabel('Frequency [Hz]')
    ax.set_ylabel('Input Referenced Current Noise [A/$\sqrt{\mathrm{Hz}}$]')
    ax.set_title(f'Normal State noise for QETbias: {qetbias*1e6} $\mu$A')
    ax.tick_params(which="both", direction="in", right=True, top=True)

    return fig, ax




##########################################
# This function will soon be deprecated
# as it does not correctly account for 
# shunt resistor errors. use qetpy.IBIS
# instead.
##########################################

import numpy as np
from scipy.optimize import curve_fit
import qetpy.plotting as utils
import warnings
warnings.simplefilter('default')

__all__ = ["IV"]

def _fitfunc(x, b, m):
    """
    Function to use for fitting to a straight line
    
    Parameters
    ----------
    x : array_like
        x-values of the data
    b : float
        y-intercept of line
    m : float
        slope of line
    
    Returns
    -------
    linfunc : array_like
        Outputted line for x with slope m and intercept b
    """
    
    linfunc = m*x + b
    return linfunc

def _findnormalinds(vb, dites, dites_err, tol=10):
    """
    Function to determine the indices of the normal data in an IV curve
    
    Parameters
    ----------
    vb : array_like
        Bias voltage, should be a 1d array or list
    dites : array_like
        The current read out by the electronics with some offset from the true current
    dites_err : array_like
        The error in the current
    tol : float, optional
        The tolerance in the reduced chi-squared for the cutoff on which points to use
            
    Returns
    -------
    normalinds : iterable
        The iterable which stores the range of the normal indices
    """
    
    end_ind = 2
    keepgoing = True
    
    while keepgoing and (end_ind < len(vb) and not np.isnan(vb[end_ind])):
        end_ind+=1
        inds = range(0,end_ind)
        
        x = curve_fit(_fitfunc, vb[inds], dites[inds], sigma=dites_err[inds], absolute_sigma=True)[0]
        red_chi2 = np.sum(((_fitfunc(vb[inds],*x)-dites[inds])/dites_err[inds])**2)/(end_ind-len(x))
        
        if red_chi2>tol:
            keepgoing = False
    
    normalinds = range(0,end_ind-1)
    
    return normalinds


class IV(object):
    """
    Class for creating the IV curve and calculating various values, such as the normal resistance, 
    the resistance of the TES, the power, etc., as well as the corresponding errors. This class supports
    data for multple bath temperatures, multiple channels, and multiple bias points.
    
    Note: If different bath temperatures have different numbers of bias points (iters), then the user
    should pad the end of the arrays with NaN so that the data can be put into an ndarray and 
    loaded into this class.
    
    Attributes
    ----------
    dites : ndarray
        The current read out by the electronics
    dites_err : ndarray
        The error in the current read out by the electronics
    vb : ndarray
        The bias voltage (vb = qet bias * rshunt)
    vb_err : ndarray
        The corresponding error in the bias voltage
    rload : scalar, ndarray
        The load resistance, this can be scalar if using the same rload for all values. If 1-dimensional, 
        then this should be the load resistance for each channel. If 2-dimensional, this should be the load
        resistance for each bath temperature and each bias point, where the shape is (ntemps, nch). If 
        3-dimensional, then this should be with shape (ntemps, nch, niters)
    rload_err : scalar, ndarray
        The corresponding error in the load resistance, should be the same type as rload
    chan_names : array_like
        Array of strings corresponding to the names of each channel in the data. Should
        have the same length as the nch axis in dites
    ioff : ndarray
        The current offset calculated from the fit, shape (ntemps, nch)
    ioff_err : ndarray
        The corresponding error in the current offset
    rfit : ndarray
        The total resistance (rnorm + rload) from the fit, shape (ntemps, nch)
    rfit_err : ndarray
        The corresponding error in the fit resistance
    rnorm : ndarray
        The normal resistance of the TES, using the fit, shape (ntemps, nch)
    rnorm_err : ndarray
        The corresponding error in the normal resistance
    ites : ndarray
        The calculated current through the TES, shape (ntemps, nch, niters)
    ites_err : ndarray
        The corresponding error in the current through the TES
    r0 : ndarray
        The calculated resistance of the TES, shape (ntemps, nch, niters)
    r0_err : ndarray
        The corresponding error in the resistance of the TES
    ptes : ndarray
        The calculated power of the TES, shape (ntemps, nch, niters)
    ptes_err : ndarray
        The corresponding error in the power of the TES
    """
    
    def __init__(self, dites, dites_err, vb, vb_err, rload, rload_err, chan_names, normalinds = None):
        """
        Initialization of the IV class object.
        
        Parameters
        ----------
        dites : ndarray
            Array of the read out current from the electronics. If 1-dimensional, should be shape (niters).
            If 2-dimensional, should be shape (nch, niters). If 3-dimensional, should be shape (ntemps, nch, niters).
            Should be the same shape as vb.
            Note: If different bath temperatures have different numbers of bias points (iters), then the user
            should pad the arrays with NaN so that the data can be put into an ndarray
        dites_err : ndarray
            The corresponding error in dites, should be same shape as dites.
        vb : ndarray
            Array of the bias voltage applied to the TES circuit (equivalent to QET bias * Rshunt).
            If 1-dimensional, should be shape (niters). If 2-dimensional, should be shape (nch, niters). 
            If 3-dimensional, should be shape (ntemps, nch, niters). Should be same shape as dites.
            Note: If different bath temperatures have different numbers of bias points (iters), then the user
            should pad the arrays with NaN so that the data can be put into an ndarray
            Should also be in the order from largest bias voltage (in magnitude) to smallest.
        vb_err : ndarray
            The corresponding error in vb (set to zeros if assuming perfect measurement), should be same 
            shape as vb.
        rload : ndarray, float
            The load resistance of the  TES circuit (equivalent to Rshunt + Rparasitic).
            If a scalar value, then the same rload is assumed for all channels, bias points, 
            and bath temperatures. If 1-dimensional, should be shape (nch,). 
            If 2-dimensional, should be shape (ntemps, nch). 
            If 3-dimensional, should be shape (ntemps, nch, niters).
        rload_err : ndarray, float
            The corresponding error in the load resistance, should be the same type/shape as rload.
        chan_names : array_like
            Array of strings corresponding to the names of each channel in the data. Should
            have the same length as the nch axis in dites
        normalinds : iterable, array_like, or NoneType, optional
            The indices of the normal resistance points. If None (default value), then the normal
            points are guessed with a simple reduced chi-squared measurement cutoff. Can also be set 
            by either an array of integers or an iterable (e.g. range(0,3)).
        
        """
        
        warnings.warn(
            "qetpy.IV class is deprecated, use qetpy.IBIS instead.\n"
            "The error propagation in qetpy.IV does not correctly take into account the shunt resistance"
            "qetpy.IV will be removed in futer versions of QETpy",
            DeprecationWarning
        )


        if len(dites.shape)==3:
            ntemps, nch, niters = dites.shape

        elif len(dites.shape)==2:
            ntemps = 1
            nch, niters = dites.shape

        elif len(dites.shape)==1:
            ntemps = 1
            nch = 1
            niters, = dites.shape

        else:
            raise ValueError("dites has too many dimensions, should be 1, 2, or 3")
        if len(chan_names) != nch:
            raise ValueError("dites has too many dimensions, should be 1, 2, or 3")
            

        # reshape arrays so the same code can be used 
        self.dites = np.reshape(dites, (ntemps, nch, niters))
        self.dites_err = np.reshape(dites_err, (ntemps, nch, niters))
        self.vb = np.reshape(vb, (ntemps, nch, niters))
        self.vb_err = np.reshape(vb_err, (ntemps, nch, niters))

        if np.isscalar(rload):
            self.rload = np.ones_like(dites)*rload
            self.rload_err = np.ones_like(dites)*rload_err

        elif rload.shape==(nch,):
            self.rload = np.tile(np.tile(rload,(niters,1)).transpose(),(ntemps,1,1))
            self.rload_err = np.tile(np.tile(rload_err,(niters,1)).transpose(),(ntemps,1,1))

        elif rload.shape==(ntemps,nch):
            self.rload = np.swapaxes(np.tile(rload.transpose(),(niters,1,1)),0,-1)
            self.rload_err = np.swapaxes(np.tile(rload_err.transpose(),(niters,1,1)),0,-1)

        elif rload.shape!=(ntemps,nch,niters):
            raise ValueError("the shape of rload doesn't match the data")
        
        self.chan_names = chan_names
        
        self.ioff = None
        self.ioff_err = None
        self.rfit = None
        self.rfit_err = None
        self.rnorm = None
        self.rnorm_err = None

        self.ites = None
        self.ites_err = None
        self.r0 = None
        self.r0_err = None
        self.ptes = None
        self.ptes_err = None

        self.normalinds = normalinds
    
    
    def calc_iv(self):
        """
        Method to calculate the IV curve for the intialized object. Calculates the power and resistance of
        each bias point, as well as saving the fit parameters from the fit to the normal points and the calculated
        normal reistance from these points.
        """

        ntemps, nch, niters = self.dites.shape
        
        ioff = np.zeros((ntemps,nch,niters))
        ioff_err = np.zeros((ntemps,nch,niters))
        rfit = np.zeros((ntemps,nch,niters))
        rfit_err = np.zeros((ntemps,nch,niters))
        rnorm = np.zeros((ntemps,nch,niters))
        rnorm_err = np.zeros((ntemps,nch,niters))
        
        for t in range(ntemps):
            for ch in range(nch):
                
                if self.normalinds is None:
                    normalinds = _findnormalinds(self.vb[t,ch], self.dites[t,ch], self.dites_err[t,ch])
                else:
                    normalinds = self.normalinds
                    
                x, xcov = curve_fit(_fitfunc, self.vb[t, ch, normalinds], self.dites[t, ch, normalinds],
                                    sigma=self.dites_err[t, ch, normalinds], absolute_sigma=True)

                jac = np.zeros((2,2))
                jac[0,0] = 1
                jac[1,1] = -1/x[1]**2

                xout = np.array([x[0],1/x[1]])
                covout = jac.dot(xcov.dot(jac.transpose()))

                ioff[t,ch] = xout[0]
                ioff_err[t,ch] = covout[0,0]**0.5

                rfit[t,ch] = xout[1]
                rfit_err[t,ch] = covout[1,1]**0.5
                
        rnorm = rfit - self.rload
        rnorm_err = (rfit_err**2 + self.rload_err**2)**0.5

        self.ites = self.dites - ioff
        self.ites_err = (self.dites_err**2.0 + ioff_err**2.0)**0.5

        self.r0 = self.vb/self.ites - self.rload
        self.r0_err = ((1/self.ites)**2*self.vb_err**2 + \
                       (self.vb/self.ites**2)**2*self.ites_err**2 + \
                       self.rload_err**2)**0.5

        self.ptes = self.ites**2 * self.r0
        self.ptes_err = ((self.vb-2*self.ites*self.rload)**2.0 * self.ites_err**2.0 + \
                         (self.ites)**2.0 * self.vb_err**2.0 + \
                         (self.ites**2.0)**2.0 * self.rload_err**2.0)**0.5
        
        self.ioff = ioff[:,:,0]
        self.ioff_err = ioff_err[:,:,0]
        self.rfit = rfit[:,:,0]
        self.rfit_err = rfit[:,:,0]
        self.rnorm = rnorm[:,:,0]
        self.rnorm_err = rnorm_err[:,:,0]

    def plot_iv(self, temps="all", chans="all", showfit=True, lgcsave=False, savepath="", savename=""):
        """
        Function to plot the IV curves for the data in an IV object.

        Parameters
        ----------
        temps : string, array_like, int, optional
            Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
            to a subset of bath temperatures, or just one
        chans : string, array_like, int, optional
            Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
            to a subset of bath temperatures, or just one
        showfit : boolean, optional
            Boolean flag to also plot the linear fit to the normal data
        lgcsave : boolean, optional
            Boolean flag to save the plot
        savepath : string, optional
            Path to save the plot to, saves it to the current directory by default
        savename : string, optional
            Name to append to the plot file name, if saving
        """
        
        fig, ax = utils.plot_iv(self, temps=temps, chans=chans, showfit=showfit, lgcsave=lgcsave, 
                        savepath=savepath, savename=savename)
        return fig, ax
        
    def plot_rv(self, temps="all", chans="all", lgcsave=False, savepath="", savename=""):
        """
        Function to plot the resistance curves for the data in an IV object.

        Parameters
        ----------
        temps : string, array_like, int, optional
            Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
            to a subset of bath temperatures, or just one
        chans : string, array_like, int, optional
            Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
            to a subset of bath temperatures, or just one
        lgcsave : boolean, optional
            Boolean flag to save the plot
        savepath : string, optional
            Path to save the plot to, saves it to the current directory by default
        savename : string, optional
            Name to append to the plot file name, if saving
        """
        
        fig, ax = utils.plot_rv(self, temps=temps, chans=chans, lgcsave=lgcsave, 
                        savepath=savepath, savename=savename)
        return fig, ax
        
    def plot_pv(self, temps="all", chans="all", lgcsave=False, savepath="", savename=""):
        """
        Function to plot the power curves for the data in an IV object.

        Parameters
        ----------
        temps : string, array_like, int, optional
            Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
            to a subset of bath temperatures, or just one
        chans : string, array_like, int, optional
            Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
            to a subset of bath temperatures, or just one
        lgcsave : boolean, optional
            Boolean flag to save the plot
        savepath : string, optional
            Path to save the plot to, saves it to the current directory by default
        savename : string, optional
            Name to append to the plot file name, if saving
        """
        
        fig, ax = utils.plot_pv(self, temps=temps, chans=chans, lgcsave=lgcsave, 
                        savepath=savepath, savename=savename)
        return fig, ax
        
    def plot_all_curves(self, temps="all", chans="all", showfit=True, lgcsave=False, savepath="", savename=""):
        """
        Function to plot the IV, resistance, and power curves for the data in an IV object.

        Parameters
        ----------
        temps : string, array_like, int, optional
            Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
            to a subset of bath temperatures, or just one
        chans : string, array_like, int, optional
            Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
            to a subset of bath temperatures, or just one
        showfit : boolean, optional
            Boolean flag to also plot the linear fit to the normal data
        lgcsave : boolean, optional
            Boolean flag to save the plot
        savepath : string, optional
            Path to save the plot to, saves it to the current directory by default
        savename : string, optional
            Name to append to the plot file name, if saving
        """
        
        utils.plot_all_curves(self, temps=temps, chans=chans, showfit=showfit, lgcsave=lgcsave, 
                                savepath=savepath, savename=savename)








import numpy as np
import random
from qetpy import ofamp
from qetpy.utils import make_template
from astropy.stats import sigma_clip
from scipy import stats, optimize
from scipy.stats import skew
import warnings
import matplotlib.pyplot as plt


__all__ = [
    "removeoutliers",
    "iterstat",
    "itercov",
    "IterCut",
    "autocuts",
    "get_muon_cut",
]


def removeoutliers(x, maxiter=20, skewtarget=0.05):
    """
    Function to return indices of inlying points, removing points
    by minimizing the skewness.

    Parameters
    ----------
    x : ndarray
        Array of real-valued variables from which to remove outliers.
    maxiter : float, optional
        Maximum number of iterations to continue to minimize skewness.
        Default is 20.
    skewtarget : float, optional
        Desired residual skewness of distribution. Default is 0.05.

    Returns
    -------
    inds : ndarray
        Boolean indices indicating which values to select/reject, same
        length as `x`.

    """

    i=1
    inds=(x != np.inf)
    sk=skew(x[inds])
    while(sk > skewtarget):
        dmed=x-np.median(x[inds])
        dist=np.min([abs(min(dmed)),abs(max(dmed))])
        inds=inds & (abs(dmed) < dist)
        sk=skew(x[inds])
        if(i > maxiter):
            break
        i+=1

    return inds

class _UnbiasedEstimators(object):
    """
    Helper class for calculating the unbiased estimators of a 1D normal
    distribution that has been truncated at specified bounds.

    Attributes
    ----------
    mu0 : float
        The biased estimator of the mean of the inputted and truncated
        data.
    std0 : float
        The biased estimator of the standard deviation of the inputted
        and truncated data.
    mu : float
        The unbiased estimator of the mean of the inputted and
        truncated data.
    std : float
        The unbiased estimator of the standard deviation of the
        inputted and truncated data.

    """
    
    def __init__(self, x, lwrbnd, uprbnd):
        """
        Initialization of the `_UnbiasedEstimators` helper class

        Parameters
        ----------
        x : ndarray
            A 1D array of data that has been truncated, for which the
            unbiased estimators will be calculated.
        lwrbnd : float
            The lower bound of the truncation of the distribution.
        uprbnd : float
            The upper bound of the truncation of the distribution.

        """

        inds = (np.asarray(x) >= lwrbnd) & (np.asarray(x) <=uprbnd)

        self._lwrbnd = lwrbnd
        self._uprbnd = uprbnd

        self._x = x[inds] # make sure data is between the specified bounds
        self._sumx = np.sum(self._x)
        self._sumx2 = np.sum(self._x**2)
        self._lenx = len(self._x)

        self.mu0 = np.mean(self._x)
        self.std0 = np.std(self._x)

        self._calc_unbiased_estimators()

    def _equations(self, p):
        """
        Helper method for calculating the system of equations that will
        be numerically solved for find the unbiased estimators.

        Parameters
        ----------
        p : tuple
            A tuple of length 2 containing the current estimated values
            of the unbiased estimators: (mu, std).

        Returns
        -------
        (mu_eqn, std_eqn) : tuple
            A tuple containing the two equations that will be solved to
            give the unbiased estimators of the mean and standard
            deviation of the data.

        """

        mu, std = p

        pdf_lwr = stats.norm.pdf(self._lwrbnd, loc=mu, scale=std)
        pdf_upr = stats.norm.pdf(self._uprbnd, loc=mu, scale=std)

        cdf_lwr = stats.norm.cdf(self._lwrbnd, loc=mu, scale=std)
        cdf_upr = stats.norm.cdf(self._uprbnd, loc=mu, scale=std)

        mu_eqn = self._sumx - self._lenx * mu
        # term due to truncation
        mu_eqn += self._lenx / (cdf_upr - cdf_lwr) * (pdf_upr - pdf_lwr)

        std_eqn = (
            self._sumx2 - 2 * mu * self._sumx
        ) + (
            self._lenx * mu**2 - self._lenx * std**2
        )
        # term due to truncation
        std_eqn += self._lenx * std**2 / (cdf_upr - cdf_lwr) * (
            (self._uprbnd - mu) * pdf_upr - (self._lwrbnd - mu) * pdf_lwr
        )

        return (mu_eqn, std_eqn)

    def _calc_unbiased_estimators(self):
        """
        Method for calculating the unbiased estimators of the truncated
        distribution.

        """

        self.mu, self.std = optimize.fsolve(
            self._equations,
            (self.mu0, self.std0),
        )


def iterstat(data, cut=3, precision=1000.0,
             return_unbiased_estimates=False):
    """
    Function to iteratively remove outliers based on how many standard
    deviations they are from the mean, where the mean and standard
    deviation are recalculated after each cut.

    Parameters
    ----------
    data : ndarray
        Array of data that we want to remove outliers from.
    cut : float, optional
        Number of standard deviations from the mean to be used for
        outlier rejection
    precision : float, optional
        Threshold for change in mean or standard deviation such that we
        stop iterating. The threshold is determined by
        np.std(data)/precision. This means that a higher number for
        precision means a lower threshold (i.e. more iterations).
    return_unbiased_estimates : bool, optional
        Boolean flag for whether or not to return the biased or
        unbiased estimates of the mean and standard deviation of the
        data. Default is False.

    Returns
    -------
    datamean : float
        Mean of the data after outliers have been removed.
    datastd : float
        Standard deviation of the data after outliers have been
        removed.
    datamask : ndarray
        Boolean array indicating which values to keep or reject in
        data, same length as data.

    """

    stdcutoff = np.std(data)/precision

    meanlast = np.mean(data)
    stdlast = np.std(data)

    nstable = 0
    keepgoing = True

    while keepgoing:
        mask = abs(data - meanlast) < cut*stdlast
        if sum(mask) <=1:
            warnings.warn(
                "The number of events passing iterative cut via iterstat is <= 1. "
                "Iteration not converging properly. Returning simple mean and std. "
                "No data will be cut."
            )
            meanthis = np.mean(data)
            stdthis = np.std(data)
            mask = np.ones(len(data),dtype=bool)
            return meanthis, stdthis, mask

        meanthis = np.mean(data[mask])
        stdthis = np.std(data[mask])

        if (
            abs(meanthis - meanlast) > stdcutoff
        ) or (
            abs(stdthis - stdlast) > stdcutoff
        ):
            nstable = 0
        else:
            nstable = nstable + 1
        if nstable >= 3:
            keepgoing = False

        meanlast = meanthis
        stdlast = stdthis

    if return_unbiased_estimates:
        unb = _UnbiasedEstimators(
            data[mask],
            meanthis - cut * stdthis,
            meanthis + cut * stdthis,
        )
        return unb.mu, unb.std, mask

    return meanthis, stdthis, mask

def itercov(*args, nsigma=2.75, threshold=None, maxiter=15,
            frac_err=1e-3):
    """
    Function for iteratively determining the estimated covariance
    matrix of a multidimensional normal distribution.

    Parameters
    ----------
    args : array_like
        The data to be iteratively cut on. Can be inputted as a single
        N-by-M array or as M arguments that are 1D vectors of length N,
        where N is the number of data points and M is the number of
        dimensions.
    nsigma : float, optional
        The number of sigma that defines that maximum chi-squared each
        data point must be below. Default is 2.75.
    threshold : float, NoneType, optional
        The threshold to cut data. If left as None, this is set to the
        larger of 3 sigma or the number of sigma such that 95% of the
        data is kept if the data is normal.
    maxiter : int, optional
        The maximum number of iterations to perform when cutting.
        Default is 15.
    frac_err : float, optional
        The fractional error allowed before stopping the iterations.
        Default is 1e-3.

    Returns
    -------
    datamean : ndarray
        The estimated mean of the data points, after iteratively
        cutting outliers.
    datacov : ndarray
        The estimated covariance of the data points, after iteratively
        cutting outliers.
    datamask : ndarray
        The boolean mask of the original data that specifies which data
        points were kept.

    Raises
    ------
    ValueError
        If the shape of the data does not match the two options
            specified by `args`.
        If the data inputted is 1-dimensional.

    """

    datashape = np.shape(args)

    if len(datashape) > 2:
        ndim = datashape[-1]
        nevts = datashape[1]
        data = args[0].T
    elif len(datashape) == 2:
        ndim = datashape[0]
        nevts = datashape[-1]
        data = np.stack(args, axis=1).T
    else:
        raise ValueError("Shape of data is inconsistent.")

    if data.shape[0] == 1:
        raise ValueError(
            "The inputted data is 1-dimensional, use qetpy.cut.iterstat instead."
        )

    if threshold is None:
        sigma2 = stats.chi2.ppf(0.95**(1 / ndim), 1)**0.5
        threshold = np.max([3, sigma2])

    mean_last = np.mean(data, axis=1)
    cov_last = np.atleast_1d(np.cov(data))

    std_last = np.sqrt(np.diag(cov_last))

    err_mean = frac_err * std_last
    err_cov = frac_err * np.sum(std_last**2)

    mean_chi2 = ndim
    sig_chi2 = np.sqrt(2 * ndim)

    max_chi2 = mean_chi2 + nsigma * sig_chi2

    nstable = 0
    keepgoing = True
    jj = 0

    while keepgoing:
        delta = data - mean_last[:, np.newaxis]

        chi2 = np.sum(delta * np.dot(np.linalg.inv(cov_last), delta), axis=0)
        mask = chi2 < max_chi2

        mask = mask & np.all(
            np.abs(delta) < std_last[:, np.newaxis] * threshold,
            axis=0,
        )
        nmask = np.sum(mask)

        if nmask <= 1:
            warnings.warn(
                "The number of events passing iterative cut via itercov is <= 1. "
                "Iteration not converging properly. Returning simple mean and cov. "
                "No data will be cut."
            )
            mean_this = np.mean(data, axis=1)
            cov_this = np.atleast_1d(np.cov(data))
            mask = np.ones(nevts, dtype=bool)

            return mean_this, cov_this, mask

        mean_this = np.mean(data[:, mask], axis=1)
        cov_this = np.atleast_1d(np.cov(data[:, mask]))

        if np.any(
            np.abs(mean_this - mean_last) > err_mean
        ) or np.any(
            np.abs(cov_this - cov_last) > err_cov
        ):
            nstable = 0
        else:
            nstable += 1

        if nstable >= 2 or jj > maxiter:
            keepgoing = False

        mean_last = mean_this
        cov_last = cov_this
        std_last = np.sqrt(np.diag(cov_last))
        jj += 1

    return mean_this, cov_this, mask

def symmetrizedist(vals):
    """
    Function to symmetrize a distribution about zero. Useful for if the
    distribution of some value centers around a nonzero value, but
    should center around zero. An example of this would be when most of
    the measured slopes are nonzero, but we want the slopes with zero
    values (e.g. lots of muon tails, which we want to cut out). To do
    this, the algorithm randomly chooses points in a histogram to cut
    out until the histogram is symmetric about zero.

    Parameters
    ----------
    vals : ndarray
        A 1-d array of the values that will be symmetrized.

    Returns
    -------
    czeromeanslope : ndarray
        A boolean mask of the values that should be kept.

    """

    nvals = len(vals)
    # figure out which direction the slopes are usually
    valsmean, valsstd = iterstat(vals, cut=2, precision=10000.0)[:-1]

    # if most vals are positive, flip the sign of them so we can use
    # the same code for both negative and positive vals
    if valsmean>0.0:
        vals= vals

    # choose symmetric upper and lower bounds for histogram to make
    # the middle bin centered on zero (since we want zero mean)
    histupr=max(vals)
    histlwr=-histupr

    # specify number of bins in histogram (should be an odd number
    # so that we have the middle bin centered on zero)
    histbins=int(np.sqrt(nvals))
    if np.mod(histbins,2)==0:
        histbins+=1

    if histupr>0:
        # create histogram, get number of events in each bin and
        # where the bin edges are
        hist_num, bin_edges = np.histogram(
            vals,
            bins=histbins,
            range=(histlwr, histupr),
        )

        if len(hist_num)>2: # otherwise we cannot symmetrize the distribution
            # inititalize the cut that symmetrizes the slopes
            czeromeanvals = np.zeros(nvals, dtype=bool)
            czeromeanvals[vals>bin_edges[histbins//2]] = True

            # go through each bin and remove events until the bin number
            # is symmetric
            for ibin in range(histbins//2, histbins-1):
                cvalsinbin = np.logical_and(
                    vals<bin_edges[histbins-ibin-1],
                    vals>=bin_edges[histbins-ibin-2],
                )
                ntracesinthisbin = hist_num[histbins-ibin-2]
                ntracesinoppobin = hist_num[ibin+1]
                ntracestoremove = ntracesinthisbin-ntracesinoppobin
                if ntracestoremove>0.0:
                    cvalsinbininds = np.where(cvalsinbin)[0]
                    crandcut = np.random.choice(
                        cvalsinbininds, ntracestoremove, replace=False,
                    )
                    cvalsinbin[crandcut] = False
                # update cut to include these events
                czeromeanvals += cvalsinbin
        else:
            # don't do anything about the shape of the distrbution
            czeromeanvals = np.ones(nvals, dtype=bool)
    else:
        # don't do anything about the shape of the distrbution
        czeromeanvals = np.ones(nvals, dtype=bool)

    return czeromeanvals


class _PlotCut(object):
    """
    Helper class for storing plotting functions for use in IterCut.

    """

    def _get_plot_inds(self, cout):
        """
        Hidden function for accessing the indices that specify passing
        events and failing events given some cut on the data.

        """

        cinds_pre = self._cutinds

        cpass = self._cutinds[cout]
        npass = len(cpass)
        cfail = np.setdiff1d(cinds_pre, cpass)
        nfail = len(cfail)

        fail_plot = np.random.choice(
            cfail,
            size=self._nplot if nfail > self._nplot else nfail,
            replace=False,
        )

        pass_plot = np.random.choice(
            cpass,
            size=self._nplot if npass > self._nplot else npass,
            replace=False,
        )

        return np.sort(fail_plot), np.sort(pass_plot)

    def _plot_events(self, cout):
        """
        Hidden function for plotting passing events and failing events
        given some cut on the data.

        """

        fail_inds, pass_inds = self._get_plot_inds(cout)

        fig, axes = plt.subplots(ncols=2, sharex=True)
        time = np.arange(self._nbin) / self.fs

        for temp_trace in self.traces[fail_inds]:
            axes[0].plot(time * 1e3, temp_trace, alpha=0.5)

        for temp_trace in self.traces[pass_inds]:
            axes[1].plot(time * 1e3, temp_trace, alpha=0.5)

        axes[0].set_title('Failed Cut')
        axes[0].set_xlim(time[0] * 1e3, time[-1] * 1e3)

        axes[1].set_title('Passed Cut')
        axes[1].yaxis.set_label_position("right")
        axes[1].yaxis.tick_right()

        for ax in axes:
            ax.tick_params(
                which='both',
                direction='in',
                right=True,
                top=True,
                left=True,
            )

        fig.add_subplot(111, frameon=False)
        plt.tick_params(
            labelcolor='none',
            which='both',
            top=False,
            bottom=False,
            left=False,
            right=False,
        )
        plt.xlabel("Time [ms]")


class IterCut(_PlotCut):
    """
    Class for iteratively applying various cuts to data while being
    able to track what events were cut between steps.

    Attributes
    ----------
    traces : ndarray
        The traces that will be cut on.
    fs : float
        The digitization rate of the traces.
    cmask : ndarray
        The current data quality cut after applying the cuts before
        this. This is a protected attribute.

    """

    def __init__(self, traces, fs, plotall=False, nplot=10):
        """
        Initialization of the IterCut class object.

        Parameters
        ----------
        traces : ndarray
            The traces that will be cut on.
        fs : float
            The digitization rate of the traces.
        plotall : bool, optional
            If True, the events that pass or fail each cut will be
            plotted at each step. Default is False. To plot events that
            pass or fail specific cuts, use the `verbose` kwarg when
            applying those cuts to the data.
        nplot : int, optional
            The number of events that should be plotted from each
            set of passing events and failing events. Default is 10.

        """

        self.traces = traces
        self.fs = fs
        self._plotall = plotall
        self._nplot = nplot
        self._ntraces = len(traces)
        self._nbin = len(traces[0])
        self._cutinds = np.arange(len(traces))

    @property
    def cmask(self):
        _cmask = np.zeros(self._ntraces, dtype=bool)
        _cmask[self._cutinds] = True
        return _cmask

    @cmask.setter
    def cmask(self, value):
        raise AttributeError("cmask is a protected attribute, can't set it")

    @cmask.deleter
    def cmask(self):
        raise AttributeError("cmask is a protected attribute, can't delete it")

    def _run_algo(self, vals, outlieralgo, verbose, **kwargs):
        """
        Hidden function for running the outlier algorithm on a set of
        values.

        """

        if outlieralgo=="iterstat":
            cout = iterstat(vals, **kwargs)[2]
        elif outlieralgo=="removeoutliers": 
            cout = removeoutliers(vals, **kwargs)
        elif outlieralgo=="sigma_clip":
            array = sigma_clip(vals, axis=0, masked=False, **kwargs)
            cout = ~np.isnan(array)
        else:
            raise ValueError(
                "Unrecognized outlieralgo, must be a str "
                "of 'iterstat', 'removeoutliers', or 'sigma_clip'"
            )

        if self._plotall or verbose:
            self._plot_events(cout)

        self._cutinds = self._cutinds[cout]

    def pileupcut(self, template=None, psd=None, removemeans=False,
                  outlieralgo="iterstat", verbose=False, **kwargs):
        """
        Function to automatically cut out outliers of the optimum
        filter amplitudes of the inputted traces.

        Parameters
        ----------
        template : ndarray, NoneType, optional
            The pulse template to use for the optimum filter. If
            not passed, then a 10 us rise time and 100 us fall time
            pulse is used.
        psd : ndarray, NoneType, optional
            The two-sided PSD (units of A^2/Hz) to use for the
            optimum filter. If not passed, then all frequencies are
            weighted equally.
        removemeans : bool, optional
            Boolean flag for if the mean of each trace should be
            removed before doing the optimum filter (True) or if the
            means should not be removed (False). This is useful for
            dIdV traces, when we want to cut out pulses that have
            smaller amplitude than the dIdV overshoot. Default is
            False.
        outlieralgo : str, optional
            Which outlier algorithm to use: iterstat, removeoutliers,
            or astropy's sigma_clip. Default is "iterstat".
        verbose : bool, optional
            If True, the events that pass or fail each cut will be
            plotted at each step. Default is False. If `plotall` is
            True when initializing this class, then this will be
            ignored in favor of `plotall`.
        **kwargs
            Keyword arguments to pass to the outlier algorithm function
            call.

        Returns
        -------
        cpileup : ndarray
            Boolean array giving which indices to keep or throw out
            based on the outlier algorithm.

        """

        if template is None:
            time = np.arange(self._nbin) / self.fs
            template = make_template(time, 10e-6, 100e-6)

        if psd is None:
            psd = np.ones(self._nbin)

        temp_traces = self.traces[self._cutinds]

        if removemeans:
            mean = np.mean(temp_traces, axis=-1, keepdims=True)
            temp_traces -= mean

        ntemptraces = len(temp_traces)

        amps = np.zeros(ntemptraces)

        #do optimum filter on all traces
        for itrace in range(ntemptraces):
            amps[itrace] = ofamp(
                temp_traces[itrace], template, psd, self.fs,
            )[0]

        self._run_algo(np.abs(amps), outlieralgo, verbose, **kwargs)

        return self.cmask

    def baselinecut(self, endindex=None, outlieralgo="iterstat",
                    verbose=False, **kwargs):
        """
        Function to automatically cut out outliers of the baselines
        of the inputted traces.

        Parameters
        ----------
        endindex : int, NoneType, optional
            The end index of the trace to average up to for the
            basleine calculation. If not passed, the default value
            is half of the trace length.
        outlieralgo : str, optional
            Which outlier algorithm to use: iterstat, removeoutliers,
            or astropy's sigma_clip. Default is "iterstat".
        verbose : bool, optional
            If True, the events that pass or fail each cut will be
            plotted at each step. Default is False. If `plotall` is
            True when initializing this class, then this will be
            ignored in favor of `plotall`.
        **kwargs
            Keyword arguments to pass to the outlier algorithm function
            call.

        Returns
        -------
        cbaseline : ndarray
            Boolean array giving which indices to keep or throw out
            based on the outlier algorithm.

        """
        
        temp_traces = self.traces[self._cutinds]
        ntemptraces = len(temp_traces)

        if endindex is None:
            endindex = self._nbin // 2

        baselines = np.mean(temp_traces[..., :endindex], axis=-1)

        self._run_algo(baselines, outlieralgo, verbose, **kwargs)

        return self.cmask

    def slopecut(self, outlieralgo="iterstat", verbose=False, **kwargs):
        """
        Function to automatically cut out outliers of the slopes of the
        inputted traces. Slopes are calculated via maximum likelihood
        and use the entire trace.

        Parameters
        ----------
        outlieralgo : str, optional
            Which outlier algorithm to use: iterstat, removeoutliers,
            or astropy's sigma_clip. Default is "iterstat".
        verbose : bool, optional
            If True, the events that pass or fail each cut will be
            plotted at each step. Default is False. If `plotall` is
            True when initializing this class, then this will be
            ignored in favor of `plotall`.
        **kwargs
            Keyword arguments to pass to the outlier algorithm function
            call.

        Returns
        -------
        cslope : ndarray
            Boolean array giving which indices to keep or throw out
            based on the outlier algorithm.

        """

        temp_traces = self.traces[self._cutinds]
        ntemptraces = len(temp_traces)
        time = np.arange(self._nbin) / self.fs
        ymean = np.mean(temp_traces, axis=-1, keepdims=True)
        xmean = np.mean(time)

        slopes = np.sum(
            (time - xmean) * (temp_traces - ymean),
            axis=-1,
        ) / np.sum(
            (time - xmean)**2,
        )

        self._run_algo(slopes, outlieralgo, verbose, **kwargs)

        return self.cmask

    def chi2cut(self, template=None, psd=None, outlieralgo="iterstat",
                verbose=False, **kwargs):
        """
        Function to automatically cut out outliers of the optimum
        filter chi-squares of the inputted traces.

        Parameters
        ----------
        template : ndarray, NoneType, optional
            The pulse template to use for the optimum filter. If
            not passed, then a 10 us rise time and 100 us fall time
            pulse is used.
        psd : ndarray, NoneType, optional
            The two-sided PSD (units of A^2/Hz) to use for the
            optimum filter. If not passed, then all frequencies are
            weighted equally.
        outlieralgo : str, optional
            Which outlier algorithm to use: iterstat, removeoutliers,
            or astropy's sigma_clip. Default is "iterstat".
        verbose : bool, optional
            If True, the events that pass or fail each cut will be
            plotted at each step. Default is False. If `plotall` is
            True when initializing this class, then this will be
            ignored in favor of `plotall`.
        **kwargs
            Keyword arguments to pass to the outlier algorithm function
            call.

        Returns
        -------
        cchi2 : ndarray
            Boolean array giving which indices to keep or throw out
            based on the outlier algorithm.

        """

        if template is None:
            time = np.arange(self._nbin) / self.fs
            template = make_template(time, 10e-6, 100e-6)

        if psd is None:
            psd = np.ones(self._nbin)

        temp_traces = self.traces[self._cutinds]
        ntemptraces = len(temp_traces)

        chi2s = np.zeros(ntemptraces)

        #do optimum filter on all traces
        for itrace in range(ntemptraces):
            chi2s[itrace] = ofamp(
                temp_traces[itrace], template, psd, self.fs,
            )[-1]

        self._run_algo(chi2s, outlieralgo, verbose, **kwargs)

        return self.cmask

    def arbitrarycut(self, cutfunction, *args, outlieralgo="iterstat",
                     verbose=False, **kwargs):
        """
        Function to automatically cut out outliers of the optimum
        filter amplitudes of the inputted traces.

        Parameters
        ----------
        cutfunction : FunctionType
            A function to set cuts on. Should be able to take
            in the traces ndarray, and any other arguments can
            be passed before defining the kwargs.
        *args
            The arguments that should be passed to `cutfunction`
            beyond the traces.
        outlieralgo : str, optional
            Which outlier algorithm to use: iterstat, removeoutliers,
            or astropy's sigma_clip. Default is "iterstat".
        verbose : bool, optional
            If True, the events that pass or fail each cut will be
            plotted at each step. Default is False. If `plotall` is
            True when initializing this class, then this will be
            ignored in favor of `plotall`.
        **kwargs
            Keyword arguments to pass to the outlier algorithm function
            call.

        Returns
        -------
        cpileup : ndarray
            Boolean array giving which indices to keep or throw out
            based on the outlier algorithm.

        """

        temp_traces = self.traces[self._cutinds]
        vals = cutfunction(temp_traces, *args)

        self._run_algo(vals, outlieralgo, verbose, **kwargs)

        return self.cmask


def autocuts(traces, fs=625e3, template=None, psd=None, is_didv=False,
             outlieralgo="iterstat", lgcpileup1=True, lgcslope=True,
             lgcbaseline=True, lgcpileup2=True, lgcchi2=True, nsigpileup1=2,
             nsigslope=2, nsigbaseline=2, nsigpileup2=2, nsigchi2=3,
             **kwargs):
    """
    Function to automatically cut out bad traces based on the optimum
    filter amplitude, slope, baseline, and chi^2 of the traces.

    Parameters
    ----------
    traces : ndarray
        2-dimensional array of traces to do cuts on
    fs : float, optional
        Sample rate that the data was taken at
    is_didv : bool, optional
        Boolean flag on whether or not the trace is a dIdV curve
    outlieralgo : string, optional
        Which outlier algorithm to use. If set to "removeoutliers",
        uses the removeoutliers algorithm that removes data based on
        the skewness of the dataset. If set to "iterstat", uses the
        iterstat algorithm to remove data based on being outside a
        certain number of standard deviations from the mean. Can also
        be set to astropy's "sigma_clip".
    lgcpileup1 : boolean, optional
        Boolean value on whether or not do the pileup1 cut (this is the
        initial pileup cut that is always done whether or not we have
        dIdV data). Default is True.
    lgcslope : boolean, optional
        Boolean value on whether or not do the slope cut. Default is
        True.
    lgcbaseline : boolean, optional
        Boolean value on whether or not do the baseline cut. Default is
        True.
    lgcpileup2 : boolean, optional
        Boolean value on whether or not do the pileup2 cut (this cut is
        only done when is_didv is also True). Default is True.
    lgcchi2 : boolean, optional
        Boolean value on whether or not do the chi2 cut. Default is
        True.
    nsigpileup1 : float, optional
        If outlieralgo is "iterstat", this can be used to tune the
        number of standard deviations from the mean to cut outliers
        from the data when using iterstat on the optimum filter
        amplitudes. Default is 2.
    nsigslope : float, optional
        If outlieralgo is "iterstat", this can be used to tune the
        number of standard deviations from the mean to cut outliers
        from the data when using iterstat on the slopes. Default is 2.
    nsigbaseline : float, optional
        If outlieralgo is "iterstat", this can be used to tune the
        number of standard deviations from the mean to cut outliers
        from the data when using iterstat on the baselines. Default is
        2.
    nsigpileup2 : float, optional
        If outlieralgo is "iterstat", this can be used to tune the
        number of standard deviations from the mean to cut outliers
        from the data when using iterstat on the optimum filter
        amplitudes after the mean has been subtracted. (only used if
        is_didv is True). Default is 2.
    nsigchi2 : float, optional
        This can be used to tune the number of standard deviations
        from the mean to cut outliers from the data when using iterstat
        on the chi^2 values. Default is 3.
    **kwargs
        Placeholder kwargs for backwards compatibility.

    Returns
    -------
    ctot : ndarray
        Boolean array giving which indices to keep or throw out based
        on the autocuts algorithm.

    """

    if is_didv and 'sgfreq' in kwargs:
        warnings.warn(
            "The `sgfreq` option has been deprecated and "
            "is now ignored when is_didv is True."
        )

    Cut = IterCut(traces, fs)

    if lgcpileup1:
        kwargs = {'cut': nsigpileup1} if outlieralgo=="iterstat" else {}
        Cut.pileupcut(
            template=template,
            psd=psd,
            outlieralgo=outlieralgo,
            **kwargs,
        )

    if lgcslope:
        kwargs = {'cut': nsigslope} if outlieralgo=="iterstat" else {}
        Cut.slopecut(outlieralgo=outlieralgo, **kwargs)

    if lgcbaseline:
        kwargs = {'cut': nsigbaseline} if outlieralgo=="iterstat" else {}
        Cut.baselinecut(outlieralgo=outlieralgo, **kwargs)

    # do a pileup cut on the mean subtracted data if this is a dIdV,
    # so that we remove pulses that are smaller than the dIdV peaks
    if lgcpileup2 and is_didv:
        kwargs = {'cut': nsigpileup2} if outlieralgo=="iterstat" else {}
        Cut.pileupcut(
            template=template,
            psd=psd,
            removemeans=True,
            outlieralgo=outlieralgo,
            **kwargs,
        )

    if lgcchi2:
        kwargs = {'cut': nsigchi2} if outlieralgo=="iterstat" else {}
        Cut.chi2cut(
            template=template,
            psd=psd,
            outlieralgo=outlieralgo,
            **kwargs,
        )

    return Cut.cmask


def get_muon_cut(traces, thresh_pct=0.95, nsatbins=600):
    """
    Function to help identify saturated muons from array of time series
    traces.

    ***Traces must have POSITIVE going pulses***

    Note, for best results, only large amplitude traces should based to
    this function. The user may need to play around with the thresh_pct
    and nsatbins parameters to achive the desired result. 

    Parameters
    ----------
    traces: array
        Array of time series traces of shape (#number of traces, #bins
        per trace).
    thresh_pct: float, optional
        The percentage of the maximum amplitude that the pulse must
        remain above for nsatbins in order to be considered
        `saturated'.
    nsatbins: int, optional
        The minimum number of bins that a muon should be saturated for.

    Returns
    -------
    muon_cut: array
        Boolean array corresponding to saturated muon events

    """

    muon_cut = np.zeros(shape = len(traces), dtype = bool)
    for ii, trace in enumerate(traces):
        trace_max = np.max(trace)
        # check that the maximum value of the trace is above the threshold and
        # that the maximum is decently larger than the minimum
        peak_loc = np.argmax(trace)

        # check that the peak is saturated (this should be true for muons
        # that saturate the detector or muon that rail the amplifier) 
        if ((peak_loc + int(nsatbins)) < traces.shape[-1]):
            if (trace[peak_loc+int(nsatbins)] >= trace_max*thresh_pct):
                muon_cut[ii] = True

    return muon_cut
